'''
Created on 2016/3/24

:author: hubo
'''
from vlcp.config.config import defaultconfig
import vlcp.service.kvdb.storage as storage
import vlcp.service.kvdb.redisnotifier as redisnotifier
from vlcp.server.module import depend, Module, callAPI, api
import vlcp.utils.dataobject as dataobj
from vlcp.event.runnable import RoutineContainer
from vlcp.event.event import Event, withIndices
from time import time
from copy import deepcopy
from vlcp.event.core import QuitException, syscall_removequeue
import itertools
from vlcp.utils.dataobject import AlreadyExistsException, UniqueKeyReference,\
    MultiKeyReference, DataObjectSet, UniqueKeySet, WeakReferenceObject,\
    MultiKeySet, ReferenceObject

@withIndices()
class RetrieveRequestSend(Event):
    pass

@withIndices('id')
class RetrieveReply(Event):
    pass

def _str(b):
    if isinstance(b, str):
        return b
    elif isinstance(b, bytes):
        return b.decode('utf-8')
    else:
        return str(b)

def _str2(b):
    if isinstance(b, str):
        return b
    elif isinstance(b, bytes):
        return b.decode('utf-8')
    elif hasattr(b, 'getkey'):
        return b.getkey()
    else:
        return str(b)


class StaleResultException(Exception):
    def __init__(self, result, desc = "Result is stale"):
        Exception.__init__(desc)
        self.result = result

class _NeedMoreKeysException(Exception):
    pass

@defaultconfig
@depend(storage.KVStorage, redisnotifier.UpdateNotifier)
class ObjectDB(Module):
    """
    Abstract transaction layer for KVDB
    """
    service = True
    # Priority for object update event
    _default_objectupdatepriority = 450
    # Enable debugging mode for updater: all updaters will be called for an extra time
    # to make sure it does not crash with multiple calls
    _default_debuggingupdater = False
    def __init__(self, server):
        Module.__init__(self, server)
        self._managed_objs = {}
        self._watches = {}
        self._watchedkeys = set()
        self._requests = []
        self._transactno = 0
        self._stale = False
        self._updatekeys = set()
        self._update_version = {}
        self.apiroutine = RoutineContainer(self.scheduler)
        self.apiroutine.main = self._update
        self.routines.append(self.apiroutine)
        self.createAPI(api(self.mget, self.apiroutine),
                       api(self.get, self.apiroutine),
                       api(self.mgetonce, self.apiroutine),
                       api(self.getonce, self.apiroutine),
                       api(self.mwatch, self.apiroutine),
                       api(self.watch, self.apiroutine),
                       api(self.munwatch, self.apiroutine),
                       api(self.unwatch, self.apiroutine),
                       api(self.unwatchall, self.apiroutine),
                       api(self.transact, self.apiroutine),
                       api(self.watchlist),
                       api(self.walk, self.apiroutine)
                       )
    def load(self, container):
        self.scheduler.queue.addSubQueue(\
                self.objectupdatepriority, dataobj.DataObjectUpdateEvent.createMatcher(), 'dataobjectupdate')
        for m in callAPI(container, 'updatenotifier', 'createnotifier'):
            yield m
        self._notifier = container.retvalue
        for m in Module.load(self, container):
            yield m
        self.routines.append(self._notifier)
    def unload(self, container, force=False):
        for m in container.syscall(syscall_removequeue(self.scheduler.queue, 'dataobjectupdate')):
            yield m
        for m in Module.unload(self, container, force=force):
            yield m
    def _update(self):
        timestamp = '%012x' % (int(time() * 1000),) + '-'
        notification_matcher = self._notifier.notification_matcher(False)
        def copywithkey(obj, key):
            newobj = deepcopy(obj)
            if hasattr(newobj, 'setkey'):
                newobj.setkey(key)
            return newobj
        def getversion(obj):
            if obj is None:
                return (0, -1)
            else:
                return (getattr(obj, 'kvdb_createtime', 0), getattr(obj, 'kvdb_updateversion', 0))
        def isnewer(obj, version):
            if obj is None:
                return version[1] != -1
            else:
                return getversion(obj) > version
        request_matcher = RetrieveRequestSend.createMatcher()
        def onupdate(event, matcher):
            update_keys = self._watchedkeys.intersection([_str(k) for k in event.keys])
            self._updatekeys.update(update_keys)
            if event.extrainfo:
                for k,v in zip(event.keys, event.extrainfo):
                    k = _str(k)
                    if k in update_keys:
                        v = tuple(v)
                        oldv = self._update_version.get(k, (0, -1))
                        if oldv < v:
                            self._update_version[k] = v
            else:
                for k in event.keys:
                    try:
                        del self._update_version[_str(k)]
                    except KeyError:
                        pass
        def updateinner():
            processing_requests = []
            # New managed keys
            retrieve_list = set()
            orig_retrieve_list = set()
            retrieveonce_list = set()
            orig_retrieveonce_list = set()
            # Retrieved values are stored in update_result before merging into current storage
            update_result = {}
            # key => [(walker_func, original_keys, rid), ...]
            walkers = {}
            # Use the loop count as a revision identifier, then the valid revisions of the value
            # in update_result is a range, from the last loop count the value changed
            # (or -1 if not changed), to the last loop count the value is retrieved
            #
            # each walker can only walk on keys that shares at least one revision to ensure the
            # values are consistent. If no revision could be shared, all the keys must be retrieved
            # again to get a consistent view
            revision_min = {}
            revision_max = {}
            self._loopCount = 0
            # A request-id -> retrieve set dictionary to store the saved keys
            savelist = {}
            def updateloop():
                while (retrieve_list or self._updatekeys or self._requests):
                    watch_keys = set()
                    # Updated keys
                    update_list = set()
                    if self._loopCount >= 10 and not retrieve_list:
                        if not self._updatekeys:
                            break
                        elif self._loopCount >= 100:
                            # Too many updates, we must stop to respond
                            self._logger.warning("There are still database updates after 100 loops of mget, respond with potential inconsistent values")
                            break
                    if self._updatekeys:
                        update_list.update(self._updatekeys)
                        self._updatekeys.clear()
                    if self._requests:
                        # Processing requests
                        for r in self._requests:
                            if r[2] == 'unwatch':
                                try:
                                    for k in r[0]:
                                        s = self._watches.get(k)
                                        if s:
                                            s.discard(r[3])
                                            if not s:
                                                del self._watches[k]
                                    # Do not need to wait
                                except Exception as exc:
                                    for m in self.apiroutine.waitForSend(RetrieveReply(r[1], exception = exc)):
                                        yield m                                    
                                else:
                                    for m in self.apiroutine.waitForSend(RetrieveReply(r[1], result = None)):
                                        yield m
                            elif r[2] == 'watch':
                                retrieve_list.update(r[0])
                                orig_retrieve_list.update(r[0])
                                for k in r[0]:
                                    self._watches.setdefault(k, set()).add(r[3])
                                processing_requests.append(r)
                            elif r[2] == 'get':
                                retrieve_list.update(r[0])
                                orig_retrieve_list.update(r[0])
                                processing_requests.append(r)
                            elif r[2] == 'walk':
                                retrieve_list.update(r[0])
                                processing_requests.append(r)
                                for k,v in r[3].items():
                                    walkers.setdefault(k, []).append((v, (r[0], r[1])))
                            else:
                                retrieveonce_list.update(r[0])
                                orig_retrieveonce_list.update(r[0])
                                processing_requests.append(r)
                        del self._requests[:]
                    if retrieve_list:
                        watch_keys.update(retrieve_list)
                    # Add watch_keys to notification
                    watch_keys.difference_update(self._watchedkeys)
                    if watch_keys:
                        for k in watch_keys:
                            if k in update_result:
                                self._update_version[k] = getversion(update_result[k])
                        for m in self._notifier.add_listen(*tuple(watch_keys.difference(self._watchedkeys))):
                            yield m
                        self._watchedkeys.update(watch_keys)
                    get_list_set = update_list.union(retrieve_list.union(retrieveonce_list).difference(self._managed_objs.keys()).difference(update_result.keys()))
                    get_list = list(get_list_set)
                    if get_list:
                        try:
                            for m in callAPI(self.apiroutine, 'kvstorage', 'mget', {'keys': get_list}):
                                yield m
                        except QuitException:
                            raise
                        except Exception:
                            # Serve with cache
                            if not self._stale:
                                self._logger.warning('KVStorage retrieve failed, serve with cache', exc_info = True)
                            self._stale = True
                            # Discard all retrieved results
                            update_result.clear()
                            # Retry update later
                            self._updatekeys.update(update_list)
                            #break
                            revision_min.clear()
                            revision_max.clear()
                        else:
                            result = self.apiroutine.retvalue
                            self._stale = False
                            for k,v in zip(get_list, result):
                                if v is not None and hasattr(v, 'setkey'):
                                    v.setkey(k)
                                if k in self._watchedkeys and k not in self._update_version:
                                    self._update_version[k] = getversion(v)
                            for k,v in zip(get_list, result):
                                # Update revision information
                                revision_max[k] = self._loopCount
                                if k not in update_result:
                                    if k not in self._managed_objs:
                                        # A newly retrieved key
                                        revision_min[k] = self._loopCount
                                        continue
                                    else:
                                        old_value = self._managed_objs[k]
                                else:
                                    old_value = update_result[k]
                                # Check if the value is changed
                                if getversion(old_value) != getversion(v):
                                    revision_min[k] = self._loopCount
                                else:
                                    if k not in revision_min:
                                        revision_min[k] = -1
                            update_result.update(zip(get_list, result))
                    # All keys which should be retrieved in next loop
                    new_retrieve_list = set()
                    # Keys which should be retrieved in next loop for a single walk
                    new_retrieve_keys = set()
                    # Keys that are used in current walk will be retrieved again in next loop
                    used_keys = set()
                    # We separate the data with revisions to prevent inconsistent result
                    def create_walker(orig_key):
                        revision_range = [revision_min.get(orig_key, -1), revision_max.get(orig_key, -1)]
                        def _walk_with_revision(key):
                            if hasattr(key, 'getkey'):
                                key = key.getkey()
                            key = _str(key)
                            if key not in self._watchedkeys:
                                # This key is not retrieved, raise a KeyError, and record this key
                                new_retrieve_keys.add(key)
                                raise KeyError('Not retrieved')
                            elif self._stale:
                                if key not in self._managed_objs:
                                    new_retrieve_keys.add(key)
                                else:
                                    used_keys.add(key)
                                return self._managed_objs.get(key)
                            elif key not in update_result and key not in self._managed_objs:
                                # This key is not retrieved, raise a KeyError, and record this key
                                new_retrieve_keys.add(key)
                                raise KeyError('Not retrieved')
                            # Check revision
                            current_revision = (
                                max(revision_min.get(key, -1), revision_range[0]),
                                min(revision_max.get(key, -1), revision_range[1])
                            )
                            if current_revision[1] < current_revision[0]:
                                # revisions cannot match
                                used_keys.add(key)
                                new_retrieve_keys.add(key)
                                raise KeyError('Not retrieved')
                            else:
                                # update revision range
                                revision_range[:] = current_revision
                            if key in update_result:
                                used_keys.add(key)
                                return update_result[key]
                            else:
                                used_keys.add(key)
                                return self._managed_objs[key]
                        return _walk_with_revision
                    walker_set = set()
                    def default_walker(key, obj, walk):
                        if key in walker_set:
                            return
                        else:
                            walker_set.add(key)
                        if hasattr(obj, 'kvdb_retrievelist'):
                            rl = obj.kvdb_retrievelist()
                            for k in rl:
                                try:
                                    newobj = walk(k)
                                except KeyError:
                                    pass
                                else:
                                    if newobj is not None:
                                        default_walker(k, newobj, walk)
                    for k in orig_retrieve_list:
                        v = update_result.get(k)
                        if v is not None:
                            new_retrieve_keys.clear()
                            used_keys.clear()
                            default_walker(k, v, create_walker(k))
                            if new_retrieve_keys:
                                new_retrieve_list.update(new_retrieve_keys)
                                self._updatekeys.update(used_keys)
                                self._updatekeys.add(k)
                    savelist.clear()
                    for k,ws in walkers.items():
                        # k: the walker key
                        # ws: list of [walker_func, (request_original_keys, rid)]
                        # Retry every walker, starts with k, with the value of v
                        if k in update_result:
                            # The value is newly retrieved
                            v = update_result.get(k)
                        else:
                            # Use the stored value
                            v = self._managed_objs.get(k)
                        if ws:
                            for w,r in list(ws):
                                # w: walker_func
                                # r: (request_original_keys, rid)
                                # Custom walker
                                def save(key):
                                    if hasattr(key, 'getkey'):
                                        key = key.getkey()
                                    key = _str(key)
                                    if key != k and key not in used_keys:
                                        raise ValueError('Cannot save a key without walk')
                                    savelist.setdefault(r[1], set()).add(key)
                                try:
                                    new_retrieve_keys.clear()
                                    used_keys.clear()
                                    w(k, v, create_walker(k), save)
                                except Exception as exc:
                                    # if one walker failed, the whole request is failed, remove all walkers
                                    self._logger.warning("A walker raises an exception which rolls back the whole walk process. "
                                                         "walker = %r, start key = %r, new_retrieve_keys = %r, used_keys = %r",
                                                         w, k, new_retrieve_keys, used_keys, exc_info=True)
                                    for orig_k in r[0]:
                                        if orig_k in walkers:
                                            walkers[orig_k][:] = [(w0, r0) for w0,r0 in walkers[orig_k] if r0[1] != r[1]]
                                    processing_requests[:] = [r0 for r0 in processing_requests if r0[1] != r[1]]
                                    savelist.pop(r[1])
                                    for m in self.apiroutine.waitForSend(RetrieveReply(r[1], exception = exc)):
                                        yield m
                                else:
                                    if new_retrieve_keys:
                                        new_retrieve_list.update(new_retrieve_keys)
                                        self._updatekeys.update(used_keys)
                                        self._updatekeys.add(k)
                    for save in savelist.values():
                        for k in save:
                            v = update_result.get(k)
                            if v is not None:
                                # If we retrieved a new value, we should also retrieved the references
                                # from this value
                                new_retrieve_keys.clear()
                                used_keys.clear()
                                default_walker(k, v, create_walker(k))
                                if new_retrieve_keys:
                                    new_retrieve_list.update(new_retrieve_keys)
                                    self._updatekeys.update(used_keys)
                                    self._updatekeys.add(k)                            
                    retrieve_list.clear()
                    retrieveonce_list.clear()
                    retrieve_list.update(new_retrieve_list)
                    self._loopCount += 1
                    if self._stale:
                        watch_keys = set(retrieve_list)
                        watch_keys.difference_update(self._watchedkeys)
                        if watch_keys:
                            for m in self._notifier.add_listen(*tuple(watch_keys)):
                                yield m
                            self._watchedkeys.update(watch_keys)
                        break
            while True:
                for m in self.apiroutine.withCallback(updateloop(), onupdate, notification_matcher):
                    yield m
                if self._loopCount >= 100 or self._stale:
                    break
                # If some updated result is newer than the notification version, we should wait for the notification
                should_wait = False
                for k,v in update_result.items():
                    if k in self._watchedkeys:
                        oldv = self._update_version.get(k)
                        if oldv is not None and isnewer(v, oldv):
                            should_wait = True
                            break
                if should_wait:
                    for m in self.apiroutine.waitWithTimeout(0.2, notification_matcher):
                        yield m
                    if self.apiroutine.timeout:
                        break
                    else:
                        onupdate(self.apiroutine.event, self.apiroutine.matcher)
                else:
                    break
            # Update result
            send_events = []
            self._transactno += 1
            transactid = '%s%016x' % (timestamp, self._transactno)
            update_objs = []
            for k,v in update_result.items():
                if k in self._watchedkeys:
                    if v is None:
                        oldv = self._managed_objs.get(k)
                        if oldv is not None:
                            if hasattr(oldv, 'kvdb_detach'):
                                oldv.kvdb_detach()
                                update_objs.append((k, oldv, dataobj.DataObjectUpdateEvent.DELETED))
                            else:
                                update_objs.append((k, None, dataobj.DataObjectUpdateEvent.DELETED))
                            del self._managed_objs[k]
                    else:
                        oldv = self._managed_objs.get(k)
                        if oldv is not None:
                            if oldv != v:
                                if oldv and hasattr(oldv, 'kvdb_update'):
                                    oldv.kvdb_update(v)
                                    update_objs.append((k, oldv, dataobj.DataObjectUpdateEvent.UPDATED))
                                else:
                                    if hasattr(oldv, 'kvdb_detach'):
                                        oldv.kvdb_detach()
                                    self._managed_objs[k] = v
                                    update_objs.append((k, v, dataobj.DataObjectUpdateEvent.UPDATED))
                        else:
                            self._managed_objs[k] = v
                            update_objs.append((k, v, dataobj.DataObjectUpdateEvent.UPDATED))
            for k in update_result.keys():
                v = self._managed_objs.get(k)
                if v is not None and hasattr(v, 'kvdb_retrievefinished'):
                    v.kvdb_retrievefinished(self._managed_objs)
            allkeys = tuple(k for k,_,_ in update_objs)
            send_events.extend((dataobj.DataObjectUpdateEvent(k, transactid, t, object = v, allkeys = allkeys) for k,v,t in update_objs))
            # Process requests
            for r in processing_requests:
                if r[2] == 'get':
                    objs = [self._managed_objs.get(k) for k in r[0]]
                    for k,v in zip(r[0], objs):
                        if v is not None:
                            self._watches.setdefault(k, set()).add(r[3])
                    result = [o.create_reference() if o is not None and hasattr(o, 'create_reference') else o
                              for o in objs]
                elif r[2] == 'watch':
                    result = [(v.create_reference() if hasattr(v, 'create_reference') else v)
                              if v is not None else dataobj.ReferenceObject(k)
                              for k,v in ((k,self._managed_objs.get(k)) for k in r[0])]
                elif r[2] == 'walk':
                    saved_keys = list(savelist.get(r[1], []))
                    for k in saved_keys:
                        self._watches.setdefault(k, set()).add(r[4])
                    objs = [self._managed_objs.get(k) for k in saved_keys]
                    result = (saved_keys,
                              [o.create_reference() if hasattr(o, 'create_reference') else o
                               if o is not None else dataobj.ReferenceObject(k)
                               for k,o in zip(saved_keys, objs)])
                else:
                    result = [copywithkey(update_result.get(k, self._managed_objs.get(k)), k) for k in r[0]]
                send_events.append(RetrieveReply(r[1], result = result, stale = self._stale))
            # Use DFS to remove unwatched objects
            mark_set = set()
            def dfs(k):
                if k in mark_set:
                    return
                mark_set.add(k)
                v = self._managed_objs.get(k)
                if v is not None and hasattr(v, 'kvdb_internalref'):
                    for k2 in v.kvdb_internalref():
                        dfs(k2)
            for k in self._watches.keys():
                dfs(k)
            def output_result():
                remove_keys = self._watchedkeys.difference(mark_set)
                if remove_keys:
                    self._watchedkeys.difference_update(remove_keys)
                    for m in self._notifier.remove_listen(*tuple(remove_keys)):
                        yield m
                    for k in remove_keys:
                        if k in self._managed_objs:
                            del self._managed_objs[k]
                        if k in self._update_version:
                            del self._update_version[k]
                for e in send_events:
                    for m in self.apiroutine.waitForSend(e):
                        yield m
            for m in self.apiroutine.withCallback(output_result(), onupdate):
                yield m
        while True:
            if not self._updatekeys and not self._requests:
                yield (notification_matcher, request_matcher)
                if self.apiroutine.matcher is notification_matcher:
                    onupdate(self.apiroutine.event, self.apiroutine.matcher)
            for m in updateinner():
                yield m
    def mget(self, keys, requestid, nostale = False):
        "Get multiple objects and manage them. Return references to the objects."
        keys = tuple(_str2(k) for k in keys)
        notify = not self._requests
        rid = object()
        self._requests.append((keys, rid, 'get', requestid))
        if notify:
            for m in self.apiroutine.waitForSend(RetrieveRequestSend()):
                yield m
        yield (RetrieveReply.createMatcher(rid),)
        if hasattr(self.apiroutine.event, 'exception'):
            raise self.apiroutine.event.exception
        if nostale and self.apiroutine.event.stale:
            raise StaleResultException(self.apiroutine.event.result)
        self.apiroutine.retvalue = self.apiroutine.event.result
    def get(self, key, requestid, nostale = False):
        """
        Get an object from specified key, and manage the object.
        Return a reference to the object or None if not exists.
        """
        for m in self.mget([key], requestid, nostale):
            yield m
        self.apiroutine.retvalue = self.apiroutine.retvalue[0]
    def mgetonce(self, keys, nostale = False):
        "Get multiple objects, return copies of them. Referenced objects are not retrieved."
        keys = tuple(_str2(k) for k in keys)
        notify = not self._requests
        rid = object()
        self._requests.append((keys, rid, 'getonce'))
        if notify:
            for m in self.apiroutine.waitForSend(RetrieveRequestSend()):
                yield m
        yield (RetrieveReply.createMatcher(rid),)
        if hasattr(self.apiroutine.event, 'exception'):
            raise self.apiroutine.event.exception
        if nostale and self.apiroutine.event.stale:
            raise StaleResultException(self.apiroutine.event.result)
        self.apiroutine.retvalue = self.apiroutine.event.result
    def getonce(self, key, nostale = False):
        "Get a object without manage it. Return a copy of the object, or None if not exists. Referenced objects are not retrieved."
        for m in self.mgetonce([key], nostale):
            yield m
        self.apiroutine.retvalue = self.apiroutine.retvalue[0]
    def watch(self, key, requestid, nostale = False):
        """
        Try to find an object and return a reference. Use ``reference.isdeleted()`` to test
        whether the object exists.
        Use ``reference.wait(container)`` to wait for the object to be existed.
        """
        for m in self.mwatch([key], requestid, nostale):
            yield m
        self.apiroutine.retvalue = self.apiroutine.retvalue[0]
    def mwatch(self, keys, requestid, nostale = False):
        "Try to return all the references, see ``watch()``"
        keys = tuple(_str2(k) for k in keys)
        notify = not self._requests
        rid = object()
        self._requests.append(keys, rid, 'watch', requestid)
        if notify:
            for m in self.apiroutine.waitForSend(RetrieveRequestSend()):
                yield m
        yield (RetrieveReply.createMatcher(rid),)
        if hasattr(self.apiroutine.event, 'exception'):
            raise self.apiroutine.event.exception
        if nostale and self.apiroutine.event.stale:
            raise StaleResultException(self.apiroutine.event.result)
        self.apiroutine.retvalue = self.apiroutine.event.result
    def unwatch(self, key, requestid):
        "Cancel management of a key"
        for m in self.munwatch([key], requestid):
            yield m
        self.apiroutine.retvalue = None
    def unwatchall(self, requestid):
        "Cancel management for all keys that are managed by requestid"
        keys = [k for k,v in self._watches.items() if requestid in v]
        for m in self.munwatch(keys, requestid):
            yield m
    def munwatch(self, keys, requestid):
        "Cancel management of keys"
        keys = tuple(_str2(k) for k in keys)
        notify = not self._requests
        rid = object()
        self._requests.append((keys, rid, 'unwatch', requestid))
        if notify:
            for m in self.apiroutine.waitForSend(RetrieveRequestSend()):
                yield m
        yield (RetrieveReply.createMatcher(rid),)
        if hasattr(self.apiroutine.event, 'exception'):
            raise self.apiroutine.event.exception
        self.apiroutine.retvalue = None
    def transact(self, keys, updater, withtime = False):
        """
        Try to update keys in a transact, with an ``updater(keys, values)``,
        which returns ``(updated_keys, updated_values)``.
        
        The updater may be called more than once. If ``withtime = True``,
        the updater should take three parameters:
        ``(keys, values, timestamp)`` with timestamp as the server time
        """
        keys = tuple(_str2(k) for k in keys)
        updated_ref = [None, None]
        extra_keys = []
        extra_key_set = []
        auto_remove_keys = set()
        orig_len = len(keys)
        def updater_with_key(keys, values, timestamp):
            # Automatically manage extra keys
            remove_uniquekeys = []
            remove_multikeys = []
            update_uniquekeys = []
            update_multikeys = []
            keystart = orig_len + len(auto_remove_keys)
            for v in values[:keystart]:
                if v is not None:
                    if hasattr(v, 'kvdb_uniquekeys'):
                        remove_uniquekeys.extend((k,v.create_weakreference()) for k in v.kvdb_uniquekeys())
                    if hasattr(v, 'kvdb_multikeys'):
                        remove_multikeys.extend((k,v.create_weakreference()) for k in v.kvdb_multikeys())
            if self.debuggingupdater:
                # Updater may be called more than once, ensure that this updater does not crash
                # on multiple calls
                kc = keys[:orig_len]
                vc = [v.clone_instance() if v is not None and hasattr(v, 'clone_instance') else deepcopy(v) for v in values[:orig_len]]
                if withtime:
                    updated_keys, updated_values = updater(kc, vc, timestamp)
                else:
                    updated_keys, updated_values = updater(kc, vc)
            if withtime:
                updated_keys, updated_values = updater(keys[:orig_len], values[:orig_len], timestamp)
            else:
                updated_keys, updated_values = updater(keys[:orig_len], values[:orig_len])
            for v in updated_values:
                if v is not None:
                    if hasattr(v, 'kvdb_uniquekeys'):
                        update_uniquekeys.extend((k,v.create_weakreference()) for k in v.kvdb_uniquekeys())
                    if hasattr(v, 'kvdb_multikeys'):
                        update_multikeys.extend((k,v.create_weakreference()) for k in v.kvdb_multikeys())
            extrakeysdict = dict(zip(keys[keystart:keystart + len(extra_keys)], values[keystart:keystart + len(extra_keys)]))
            extrakeysetdict = dict(zip(keys[keystart + len(extra_keys):keystart + len(extra_keys) + len(extra_key_set)],
                                       values[keystart + len(extra_keys):keystart + len(extra_keys) + len(extra_key_set)]))
            tempdict = {}
            old_values = dict(zip(keys, values))
            updated_keyset = set(updated_keys)
            try:
                append_remove = set()
                autoremove_keys = set()
                # Use DFS to find auto remove keys
                def dfs(k):
                    if k in autoremove_keys:
                        return
                    autoremove_keys.add(k)
                    if k not in old_values:
                        append_remove.add(k)
                    else:
                        oldv = old_values[k]
                        if oldv is not None and hasattr(oldv, 'kvdb_autoremove'):
                            for k2 in oldv.kvdb_autoremove():
                                dfs(k2)
                for k,v in zip(updated_keys, updated_values):
                    if v is None:
                        dfs(k)
                if append_remove:
                    raise _NeedMoreKeysException()
                for k,v in remove_uniquekeys:
                    if v.getkey() not in updated_keyset and v.getkey() not in auto_remove_keys:
                        # This key is not updated, keep the indices untouched
                        continue
                    if k not in extrakeysdict:
                        raise _NeedMoreKeysException()
                    elif extrakeysdict[k] is not None and extrakeysdict[k].ref.getkey() == v.getkey():
                        # If the unique key does not reference to the correct object
                        # there may be an error, but we ignore this.
                        # Save in a temporary dictionary. We may restore it later.
                        tempdict[k] = extrakeysdict[k]
                        extrakeysdict[k] = None
                        setkey = UniqueKeyReference.get_keyset_from_key(k)
                        if setkey not in extrakeysetdict:
                            raise _NeedMoreKeysException()
                        else:
                            ks = extrakeysetdict[setkey]
                            if ks is None:
                                ks = UniqueKeySet.create_from_key(setkey)
                                extrakeysetdict[setkey] = ks
                            ks.set.dataset().discard(WeakReferenceObject(k))
                for k,v in remove_multikeys:
                    if v.getkey() not in updated_keyset and v.getkey() not in auto_remove_keys:
                        # This key is not updated, keep the indices untouched
                        continue
                    if k not in extrakeysdict:
                        raise _NeedMoreKeysException()
                    else:
                        mk = extrakeysdict[k]
                        if mk is not None:
                            mk.set.dataset().discard(v)
                            if not mk.set.dataset():
                                tempdict[k] = extrakeysdict[k]
                                extrakeysdict[k] = None
                                setkey = MultiKeyReference.get_keyset_from_key(k)
                                if setkey not in extrakeysetdict:
                                    raise _NeedMoreKeysException()
                                else:
                                    ks = extrakeysetdict[setkey]
                                    if ks is None:
                                        ks = MultiKeySet.create_from_key(setkey)
                                        extrakeysetdict[setkey] = ks
                                    ks.set.dataset().discard(WeakReferenceObject(k))
                for k,v in update_uniquekeys:
                    if k not in extrakeysdict:
                        raise _NeedMoreKeysException()
                    elif extrakeysdict[k] is not None and extrakeysdict[k].ref.getkey() != v.getkey():
                        raise AlreadyExistsException('Unique key conflict for %r and %r, with key %r' % \
                                                     (extrakeysdict[k].ref.getkey(), v.getkey(), k))
                    elif extrakeysdict[k] is None:
                        lv = tempdict.get(k, None)
                        if lv is not None and lv.ref.getkey() == v.getkey():
                            # Restore this value
                            nv = lv
                        else:
                            nv = UniqueKeyReference.create_from_key(k)
                            nv.ref = ReferenceObject(v.getkey())
                        extrakeysdict[k] = nv
                        setkey = UniqueKeyReference.get_keyset_from_key(k)
                        if setkey not in extrakeysetdict:
                            raise _NeedMoreKeysException()
                        else:
                            ks = extrakeysetdict[setkey]
                            if ks is None:
                                ks = UniqueKeySet.create_from_key(setkey)
                                extrakeysetdict[setkey] = ks
                            ks.set.dataset().add(nv.create_weakreference())
                for k,v in update_multikeys:
                    if k not in extrakeysdict:
                        raise _NeedMoreKeysException()
                    else:
                        mk = extrakeysdict[k]
                        if mk is None:
                            mk = tempdict.get(k, None)
                            if mk is None:
                                mk = MultiKeyReference.create_from_key(k)
                                mk.set = DataObjectSet()
                            setkey = MultiKeyReference.get_keyset_from_key(k)
                            if setkey not in extrakeysetdict:
                                raise _NeedMoreKeysException()
                            else:
                                ks = extrakeysetdict[setkey]
                                if ks is None:
                                    ks = MultiKeySet.create_from_key(setkey)
                                    extrakeysetdict[setkey] = ks
                                ks.set.dataset().add(mk.create_weakreference())
                        mk.set.dataset().add(v)
                        extrakeysdict[k] = mk
            except _NeedMoreKeysException:
                # Prepare the keys
                extra_keys[:] = list(set(itertools.chain((k for k,v in remove_uniquekeys if v.getkey() in updated_keyset or v.getkey() in autoremove_keys),
                                                         (k for k,v in remove_multikeys if v.getkey() in updated_keyset or v.getkey() in autoremove_keys),
                                                         (k for k,_ in update_uniquekeys),
                                                         (k for k,_ in update_multikeys))))
                extra_key_set[:] = list(set(itertools.chain((UniqueKeyReference.get_keyset_from_key(k) for k,v in remove_uniquekeys if v.getkey() in updated_keyset or v.getkey() in autoremove_keys),
                                                         (MultiKeyReference.get_keyset_from_key(k) for k,v in remove_multikeys if v.getkey() in updated_keyset or v.getkey() in autoremove_keys),
                                                         (UniqueKeyReference.get_keyset_from_key(k) for k,_ in update_uniquekeys),
                                                         (MultiKeyReference.get_keyset_from_key(k) for k,_ in update_multikeys))))
                auto_remove_keys.clear()
                auto_remove_keys.update(autoremove_keys.difference(keys[:orig_len])
                                                          .difference(extra_keys)
                                                          .difference(extra_key_set))
                raise
            else:
                extrakeys_list = list(extrakeysdict.items())
                extrakeyset_list = list(extrakeysetdict.items())
                autoremove_list = list(autoremove_keys.difference(updated_keys)
                                                      .difference(extrakeysdict.keys())
                                                      .difference(extrakeysetdict.keys()))
                return (tuple(itertools.chain(updated_keys,
                                              (k for k,_ in extrakeys_list),
                                              (k for k,_ in extrakeyset_list),
                                              autoremove_list)),
                        tuple(itertools.chain(updated_values,
                                               (v for _,v in extrakeys_list),
                                               (v for _,v in extrakeyset_list),
                                               [None] * len(autoremove_list))))
                        
        def object_updater(keys, values, timestamp):
            old_version = {}
            for k, v in zip(keys, values):
                if v is not None and hasattr(v, 'setkey'):
                    v.setkey(k)
                if v is not None and hasattr(v, 'kvdb_createtime'):
                    old_version[k] = (getattr(v, 'kvdb_createtime'), getattr(v, 'kvdb_updateversion', 1))
            updated_keys, updated_values = updater_with_key(keys, values, timestamp)
            updated_ref[0] = tuple(updated_keys)
            new_version = []
            for k,v in zip(updated_keys, updated_values):
                if v is None:
                    new_version.append((timestamp, -1))
                elif k in old_version:
                    ov = old_version[k]
                    setattr(v, 'kvdb_createtime', ov[0])
                    setattr(v, 'kvdb_updateversion', ov[1] + 1)
                    new_version.append((ov[0], ov[1] + 1))
                else:
                    setattr(v, 'kvdb_createtime', timestamp)
                    setattr(v, 'kvdb_updateversion', 1)
                    new_version.append((timestamp, 1))
            updated_ref[1] = new_version
            return (updated_keys, updated_values)
        while True:
            try:
                for m in callAPI(self.apiroutine, 'kvstorage', 'updateallwithtime',
                                 {'keys': keys + tuple(auto_remove_keys) + \
                                         tuple(extra_keys) + tuple(extra_key_set),
                                         'updater': object_updater}):
                    yield m
            except _NeedMoreKeysException:
                pass
            else:
                break
        # Short cut update notification
        update_keys = self._watchedkeys.intersection(updated_ref[0])
        self._updatekeys.update(update_keys)
        for k,v in zip(updated_ref[0], updated_ref[1]):
            k = _str(k)
            if k in update_keys:
                v = tuple(v)
                oldv = self._update_version.get(k, (0, -1))
                if oldv < v:
                    self._update_version[k] = v
        for m in self.apiroutine.waitForSend(RetrieveRequestSend()):
            yield m
        for m in self._notifier.publish(updated_ref[0], updated_ref[1]):
            yield m
    def watchlist(self, requestid = None):
        """
        Return a dictionary whose keys are database keys, and values are lists of request ids.
        Optionally filtered by request id
        """
        return dict((k,list(v)) for k,v in self._watches.items() if requestid is None or requestid in v)
    def walk(self, keys, walkerdict, requestid, nostale = False):
        """
        Recursively retrieve keys with customized functions.
        walkerdict is a dictionary ``key->walker(key, obj, walk, save)``.
        """
        keys = tuple(_str2(k) for k in keys)
        notify = not self._requests
        rid = object()
        self._requests.append((keys, rid, 'walk', dict(walkerdict), requestid))
        if notify:
            for m in self.apiroutine.waitForSend(RetrieveRequestSend()):
                yield m
        yield (RetrieveReply.createMatcher(rid),)
        if hasattr(self.apiroutine.event, 'exception'):
            raise self.apiroutine.event.exception
        if nostale and self.apiroutine.event.stale:
            raise StaleResultException(self.apiroutine.event.result)
        self.apiroutine.retvalue = self.apiroutine.event.result
        


'''
Created on 2015/11/9

:author: hubo
'''
from vlcp.config import defaultconfig
from vlcp.server.module import Module, api, proxy
from vlcp.event.core import TimerEvent
from vlcp.event.runnable import RoutineContainer
from time import time
import functools

@defaultconfig
class Knowledge(Module):
    '''
    Simple KV-cache in memory. A base for other KV-DB.
    Data is automatically removed after timeout.
    Use knowledge instead of local storage in modules so data is not lost on module restarting.
    '''
    # Check current data set, remove the expired data
    _default_checkinterval = 300
    service = True
    def __init__(self, server):
        '''
        Constructor
        '''
        Module.__init__(self, server)
        self.timeoutroutine = RoutineContainer(self.scheduler)
        self.timeoutroutine.main = self._timeout
        self.db = {}
        self.routines.append(self.timeoutroutine)
        self.createAPI(api(self.get),
                       api(self.set),
                       api(self.delete),
                       api(self.mget),
                       api(self.mset),
                       api(self.update),
                       api(self.mupdate),
                       api(self.updateall),
                       api(self.updateallwithtime))
    def _timeout(self):
        th = self.scheduler.setTimer(self.checkinterval, self.checkinterval)
        try:
            tm = TimerEvent.createMatcher(th)
            while True:
                yield (tm,)
                t = time()
                timeouts = [k for k,v in self.db.items() if v[1] is not None and v[1] < t]
                for k in timeouts:
                    del self.db[k]
        finally:
            self.scheduler.cancelTimer(th)
    def _get(self, key, currtime):
        v = self.db.get(key)
        if v is not None:
            if v[1] is not None and v[1] < currtime:
                del self.db[key]
                return None
        return v
    def _set(self, key, value, currtime, timeout):
        if timeout is not None and timeout <= 0:
            try:
                del self.db[key]
            except KeyError:
                pass
        else:
            if timeout is None:
                self.db[key] = (value, None)
            else:
                self.db[key] = (value, currtime + timeout)
    def get(self, key, timeout = None):
        "Get value from key"
        t = time()
        v = self._get(key, t)
        if v is None:
            return None
        if timeout is not None:
            self._set(key, v[0], t, timeout)
        return v[0]
    def set(self, key, value, timeout = None):
        "Set value to key, with an optional timeout"
        self._set(key, value, time(), timeout)
        return None
    def delete(self, key):
        "Delete a key"
        try:
            del self.db[key]
        except KeyError:
            pass
        return None
    def mget(self, keys):
        "Get multiple values from multiple keys"
        t = time()
        return [None if v is None else v[0] for v in (self._get(k, t) for k in keys)]
    def mset(self, kvpairs, timeout = None):
        "Set multiple values on multiple keys"
        d = kvpairs
        if hasattr(d, 'items'):
            d = d.items()
        if timeout is not None and timeout <= 0:
            for k,_ in d:
                self.delete(k)
        else:
            if timeout is None:
                t = None
            else:
                t = time() + timeout
            self.db.update(((k, (v, t)) for k,v in d))
        return None
    def update(self, key, updater, timeout = None):
        '''
        Update in-place with a custom function
        
        :param key: key to update
        
        :param updater: ``func(k,v)``, should return a new value to update, or return None to delete
        
        :param timeout: new timeout
        
        :returns: the updated value, or None if deleted
        '''
        return self._update(key, updater, time(), timeout)
    def _update(self, key, updater, currtime, timeout):
        v = self._get(key, currtime)
        if v is not None:
            v = v[0]
        newv = updater(key, v)
        if newv is not None:
            self._set(key, newv, currtime, timeout)
        else:
            self.delete(key)
        return newv
    def mupdate(self, keys, updater, timeout = None):
        "Update multiple keys in-place one by one with a custom function, see update. Either all success, or all fail."
        t = time()
        vs = [v[0] if v is not None else None for v in (self._get(k, t) for k in keys)]
        newvs = [updater(k, v) for k,v in zip(keys, vs)]
        for k,v in zip(keys, newvs):
            if v is not None:
                self._set(k, v, t, timeout)
            else:
                self.delete(k)
        return newvs
    def updateall(self, keys, updater, timeout = None):
        """
        Update multiple keys in-place, with a function ``updater(keys, values)`` which returns
        ``(updated_keys, updated_values)``.
        Either all success or all fail
        """
        t = time()
        vs = [v[0] if v is not None else None  for v in (self._get(k, t) for k in keys)]
        newkeys, newvs = updater(keys, vs)
        for k,v in zip(newkeys, newvs):
            if v is not None:
                self._set(k, v, t, timeout)
            else:
                self.delete(k)
        return (newkeys, newvs)
    def updateallwithtime(self, keys, updater, timeout = None):
        """
        Update multiple keys in-place, with a function ``updater(keys, values, timestamp)``
        which returns ``(updated_keys, updated_values)``. Either all success or all fail.
        
        Timestamp is a integer standing for current time in microseconds.
        """
        t = time()
        vs = [v[0] if v is not None else None  for v in (self._get(k, t) for k in keys)]
        newkeys, newvs = updater(keys, vs, int(t * 1000000))
        for k,v in zip(newkeys, newvs):
            if v is not None:
                self._set(k, v, t, timeout)
            else:
                self.delete(k)
        return (newkeys, newvs)

def return_self_updater(func):
    '''
    Run func, but still return v. Useful for using knowledge.update with operates like append, extend, etc.
    e.g. return_self(lambda k,v: v.append('newobj'))
    '''
    @functools.wraps(func)
    def decorator(k,v):
        func(k,v)
        return v
    return decorator

def escape_key(k):
    "Escape k, ensuring there is not a '.' in the string."
    return k.replace('+','+_').replace('.','++')

def unescape_key(k):
    "Unescape key to get '.' back"
    return k.replace('++', '.').replace('+_','+')

MemoryStorage = proxy('MemoryStorage', Knowledge)



'''
Created on 2015/6/25

:author: hubo
'''
import re
import ast
import sys
import io

class ConfigTree(object):
    """
    A basic config node. A node supports both attributes get/set and dict-like operations. When
    using dict-like interfaces, configurations in child nodes can directly be used::
    
        node['child.childconfig'] = 42
        node.child.childconfig  # 42
    """
    def __init__(self):
        pass
    def keys(self):
        """
        Return all children in this node, either sub nodes or configuration values
        """
        return self.__dict__.keys()
    def items(self):
        """
        Return `(key, value)` tuples for children in this node, either sub nodes or configuration values
        """
        return self.__dict__.items()
    def __len__(self):
        """
        Return size of children stored in this node, either sub nodes or configuration values
        """
        return len(self.__dict__)
    def config_keys(self, sortkey = False):
        """
        Return all configuration keys in this node, including configurations on children nodes.
        """
        if sortkey:
            items = sorted(self.items())
        else:
            items = self.items()
        for k,v in items:
            if isinstance(v, ConfigTree):
                for k2 in v.config_keys(sortkey):
                    yield k + '.' + k2
            else:
                yield k
    def config_items(self, sortkey = False):
        """
        Return all `(key, value)` tuples for configurations in this node, including configurations on children nodes.
        """
        if sortkey:
            items = sorted(self.items())
        else:
            items = self.items()
        for k,v in items:
            if isinstance(v, ConfigTree):
                for k2,v2 in v.config_items(sortkey):
                    yield (k + '.' + k2, v2)
            else:
                yield (k,v)
    def config_value_keys(self, sortkey = False):
        """
        Return configuration keys directly stored in this node. Configurations in child nodes are not included.
        """
        if sortkey:
            items = sorted(self.items())
        else:
            items = self.items()
        return (k for k,v in items if not isinstance(v,ConfigTree))        
    def config_value_items(self, sortkey = False):
        """
        Return `(key, value)` tuples for configuration directly stored in this node. Configurations in child nodes are not included.
        """
        if sortkey:
            items = sorted(self.items())
        else:
            items = self.items()        
        return ((k,v) for k,v in items if not isinstance(v,ConfigTree))
    def loadconfig(self, keysuffix, obj):
        """
        Copy all configurations from this node into obj
        """
        subtree = self.get(keysuffix)
        if subtree is not None and isinstance(subtree, ConfigTree):
            for k,v in subtree.items():
                if isinstance(v, ConfigTree):
                    if hasattr(obj, k) and not isinstance(getattr(obj, k), ConfigTree):
                        v.loadconfig(getattr(obj,k))
                    else:
                        setattr(obj, k, v)
                elif not hasattr(obj, k):
                    setattr(obj, k, v)
    def withconfig(self, keysuffix):
        """
        Load configurations with this decorator
        """
        def decorator(cls):
            return self.loadconfig(keysuffix, cls)
        return decorator
    def __iter__(self):
        """
        Support a dict-like iterate
        """
        return iter(self.__dict__)
    def gettree(self, key, create = False):
        """
        Get a subtree node from the key (path relative to this node)
        """
        tree, _ = self._getsubitem(key + '.tmp', create)
        return tree
    def _getsubitem(self, key, create = False):
        keylist = [k for k in key.split('.') if k != '']
        if not keylist:
            raise KeyError('Config key is empty')
        current = self
        for k in keylist[:-1]:
            if not hasattr(current, k):
                if create:
                    c2 = ConfigTree()
                    setattr(current, k, c2)
                    current = c2
                else:
                    return (None, None)
            else:
                v = getattr(current, k)
                if not isinstance(v, ConfigTree):
                    if create:
                        c2 = ConfigTree()
                        setattr(current, k, c2)
                        current = c2
                    else:
                        return (None, None)
                else:
                    current = v
        return (current, keylist[-1])
    def __setitem__(self, key, value):
        """
        Support dict-like assignment
        """
        (t, k) = self._getsubitem(key, True)
        setattr(t, k, value)
    def __delitem__(self, key):
        """
        Support dict-like deletion
        """
        (t, k) = self._getsubitem(key, False)
        if t is None:
            raise KeyError(key)
        else:
            delattr(t, k)
    def __getitem__(self, key):
        """
        Support dict-like retrieval
        """
        (t, k) = self._getsubitem(key, False)
        if t is None:
            raise KeyError(key)
        else:
            return t.__dict__[k]
    def get(self, key, defaultvalue = None):
        """
        Support dict-like get (return a default value if not found)
        """
        (t, k) = self._getsubitem(key, False)
        if t is None:
            return defaultvalue
        else:
            return t.__dict__.get(k, defaultvalue)
    def setdefault(self, key, defaultvalue = None):
        """
        Support dict-like setdefault (create if not existed)
        """
        (t, k) = self._getsubitem(key, True)
        return t.__dict__.setdefault(k, defaultvalue)
    def __contains__(self, key):
        """
        Support dict-like `in` operator
        """
        (t, k) = self._getsubitem(key, False)
        if t is None:
            return False
        else:
            return k in t.__dict__
    def clear(self):
        """
        Support dict-like clear
        """
        self.__dict__.clear()
    def todict(self):
        """
        Convert this node to a dictionary tree.
        """
        dict_entry = []
        for k,v in self.items():
            if isinstance(v, ConfigTree):
                dict_entry.append((k, v.todict()))
            else:
                dict_entry.append((k, v))
        return dict(dict_entry)

class Manager(ConfigTree):
    '''
    Configuration manager. Use the global variable `manager` to access the configuration system.
    '''
    def __init__(self):
        ConfigTree.__init__(self)
    def loadfromfile(self, filelike):
        """
        Read configurations from a file-like object, or a sequence of strings. Old values are not
        cleared, if you want to reload the configurations completely, you should call `clear()`
        before using `load*` methods.
        """
        line_format = re.compile(r'((?:[a-zA-Z][a-zA-Z0-9_]*\.)*[a-zA-Z][a-zA-Z0-9_]*)\s*=\s*')
        space = re.compile(r'\s')
        line_no = 0
        # Suport multi-line data
        line_buffer = []
        line_key = None
        last_line_no = None
        for l in filelike:
            line_no += 1
            ls = l.strip()
            # If there is a # in start, the whole line is remarked.
            # ast.literal_eval already processed any # after '='.
            if not ls or ls.startswith('#'):
                continue
            if space.match(l):
                if not line_key:
                    # First line cannot start with space
                    raise ValueError('Error format in line %d: first line cannot start with space%s\n' % (line_no, l))
                line_buffer.append(l)
            else:
                if line_key:
                    # Process buffered lines
                    try:
                        value = ast.literal_eval(''.join(line_buffer))
                    except:
                        typ, val, tb = sys.exc_info()
                        raise ValueError('Error format in line %d(%s: %s):\n%s' % (last_line_no, typ.__name__, str(val), ''.join(line_buffer)))
                    self[line_key] = value
                # First line
                m = line_format.match(l)
                if not m:
                    raise ValueError('Error format in line %d:\n%s' % (line_no, l))
                line_key = m.group(1)
                last_line_no = line_no
                del line_buffer[:]
                line_buffer.append(l[m.end():])
        if line_key:
            # Process buffered lines
            try:
                value = ast.literal_eval(''.join(line_buffer))
            except:
                typ, val, tb = sys.exc_info()
                raise ValueError('Error format in line %d(%s: %s):\n%s' % (last_line_no, typ.__name__, str(val), ''.join(line_buffer)))
            self[line_key] = value
    def loadfrom(self, path):
        """
        Read configurations from path
        """
        with open(path, 'r') as f:
            self.loadfromfile(f)
    def loadfromstr(self, string):
        """
        Read configurations from string
        """
        self.loadfromfile(string.splitlines(keepends=True))
    def save(self, sortkey = True):
        """
        Save configurations to a list of strings
        """
        return [k + '=' + repr(v) for k,v in self.config_items(sortkey)]
    def savetostr(self, sortkey = True):
        """
        Save configurations to a single string
        """
        return ''.join(k + '=' + repr(v) + '\n' for k,v in self.config_items(sortkey))
    def savetofile(self, filelike, sortkey = True):
        """
        Save configurations to a file-like object which supports `writelines`
        """
        filelike.writelines(k + '=' + repr(v) + '\n' for k,v in self.config_items(sortkey))
    def saveto(self, path, sortkey = True):
        """
        Save configurations to path
        """
        with open(path, 'w') as f:
            self.savetofile(f, sortkey)

# Global configuration manager
manager = Manager()

class Configurable(object):
    """
    Base class for a configurable object. Undefined attributes of a configurable object is mapped to
    global configurations. The attribute value of a configurable object is:
    
    1. The original attribute value if it is set on the instance or class
    
    2. The configuration value `manager[self.configkey + '.' + attrname]` if exists
    
    3. The configuration value `manager[parent.configkey + '.' + attrname]` if parent classes have
       `configkey` defined and configuration exists.
    
    4. The `_default_<attrname>` attribute value of the instance
    
    5. raises AttributeError
    
    Attributes begins with '_' is not mapped.
    
    `configkey` and `configbase` attribute should be set on this class. Usually they are set by decorators
    `@defaultconfig`, `@configbase` or `@config`
    """
    def __init__(self):
        pass
    def __getattr__(self, key):
        if key.startswith('_'):
            raise AttributeError("type object '%s' has no attribute '%s'" % (type(self).__name__, key))
        cls = type(self)
        while True:
            try:
                return manager[getattr(cls, 'configkey') + '.' + key]
            except AttributeError:
                pass
            except KeyError:
                pass
            try:
                return cls.__dict__['_default_' + key]
            except KeyError:
                pass
            parent = None
            for c in cls.__bases__:
                if issubclass(c, Configurable):
                    parent = c
            if parent is None:
                break
            cls = parent
        raise AttributeError("type object '%s' has no attribute '%s'" % (type(self).__name__, key))
    @classmethod
    def getConfigurableParent(cls):
        """
        Return the parent from which this class inherits configurations
        """
        for p in cls.__bases__:
            if isinstance(p, Configurable) and p is not Configurable:
                return p
        return None
    @classmethod
    def getConfigRoot(cls, create = False):
        """
        Return the mapped configuration root node
        """
        try:
            return manager.gettree(getattr(cls, 'configkey'), create)
        except AttributeError:
            return None
    def config_value_keys(self, sortkey = False):
        """
        Return all mapped configuration keys for this object
        """
        ret = set()
        cls = type(self)
        while True:
            root = cls.getConfigRoot()
            if root:
                ret = ret.union(set(root.config_value_keys()))
            parent = None
            for c in cls.__bases__:
                if issubclass(c, Configurable):
                    parent = c
            if parent is None:
                break
            cls = parent
        if sortkey:
            return sorted(list(ret))
        else:
            return list(ret)
        
    def config_value_items(self, sortkey = False):
        """
        Return `(key, value)` tuples for all mapped configurations for this object
        """
        return ((k, getattr(self, k)) for k in self.config_value_keys(sortkey))
            
def configbase(key):
    """
    Decorator to set this class to configuration base class. A configuration base class
    uses `<parentbase>.key.` for its configuration base, and uses `<parentbase>.key.default` for configuration mapping.
    """
    def decorator(cls):
        parent = cls.getConfigurableParent()
        if parent is None:
            parentbase = None
        else:
            parentbase = getattr(parent, 'configbase', None)
        if parentbase is None:
            base = key
        else:
            base = parentbase + '.' + key
        cls.configbase = base
        cls.configkey = base + '.default'
        return cls
    return decorator

def config(key):
    """
    Decorator to map this class directly to a configuration node. It uses `<parentbase>.key` for configuration
    base and configuration mapping.
    """
    def decorator(cls):
        parent = cls.getConfigurableParent()
        if parent is None:
            parentbase = None
        else:
            parentbase = getattr(parent, 'configbase', None)
        if parentbase is None:
            cls.configkey = key
        else:
            cls.configkey = parentbase + '.' + key
        return cls
    return decorator

def defaultconfig(cls):
    """
    Generate a default configuration mapping bases on the class name. If this class does not have a
    parent with `configbase` defined, it is set to a configuration base with
    `configbase=<lowercase-name>` and `configkey=<lowercase-name>.default`; otherwise it inherits
    `configbase` of its parent and set `configkey=<parentbase>.<lowercase-name>`
    
    Refer to :ref::`configurations` for normal rules.
    """
    parentbase = None
    for p in cls.__bases__:
        if issubclass(p, Configurable):
            parentbase = getattr(p, 'configbase', None)
            break
    if parentbase is None:
        base = cls.__name__.lower()
        cls.configbase = base
        cls.configkey = base + '.default'
    else:
        key = cls.__name__.lower()
        #=======================================================================
        # parentkeys = parentbase.split('.')
        # for pk in parentkeys:
        #     if key.endswith(pk):
        #         key = key[0:-len(pk)]
        #     elif key.startswith(pk):
        #         key = key[len(pk):]
        #=======================================================================
        cls.configkey = parentbase + "." + key
    return cls



'''
Created on 2016/12/15

:author: hubo

Refactoring from vxlancast module, to share the logic with hardware-vtep
'''
from vlcp.utils.networkmodel import VXLANEndpointSet, LogicalNetworkMap,\
    LogicalPort, LogicalPortVXLANInfo
import itertools
from vlcp.server.module import callAPI
from vlcp.utils.ethernet import ip4_addr


def lognet_vxlan_walker(prepush = True):
    """
    Return a walker function to retrieve necessary information from ObjectDB
    """
    def _walk_lognet(key, value, walk, save):
        save(key)
        if value is None:
            return
        try:
            phynet = walk(value.physicalnetwork.getkey())
        except KeyError:
            pass
        else:
            if phynet is not None and getattr(phynet, 'type') == 'vxlan':
                try:
                    vxlan_endpoint_key = VXLANEndpointSet.default_key(value.id)
                    walk(vxlan_endpoint_key)
                except KeyError:
                    pass
                else:
                    save(vxlan_endpoint_key)
                if prepush:
                    # Acquire all logical ports
                    try:
                        netmap = walk(LogicalNetworkMap.default_key(value.id))
                    except KeyError:
                        pass
                    else:
                        save(netmap.getkey())
                        for logport in netmap.ports.dataset():
                            try:
                                _ = walk(logport.getkey())
                            except KeyError:
                                pass
                            else:
                                save(logport.getkey())
                            try:
                                _, (portid,) = LogicalPort._getIndices(logport.getkey())
                                portinfokey = LogicalPortVXLANInfo.default_key(portid)
                                _ = walk(portinfokey)
                            except KeyError:
                                pass
                            else:
                                save(portinfokey)                            
    return _walk_lognet

def update_vxlaninfo(container, network_ip_dict, created_ports, removed_ports,
                ovsdb_vhost, system_id, bridge,
                allowedmigrationtime, refreshinterval):
    '''
    Do an ObjectDB transact to update all VXLAN informations
    
    :param container: Routine container
    
    :param network_ip_dict: a {logicalnetwork_id: tunnel_ip} dictionary
    
    :param created_ports: logical ports to be added, a {logicalport_id: tunnel_ip} dictionary
    
    :param removed_ports: logical ports to be removed, a {logicalport_id: tunnel_ip} dictionary
    
    :param ovsdb_vhost: identifier for the bridge, vhost name
    
    :param system_id: identifier for the bridge, OVSDB systemid
    
    :param bridge: identifier for the bridge, bridge name
    
    :param allowedmigrationtime: time allowed for port migration, secondary endpoint info will be removed
                                 after this time
    
    :param refreshinterval: refreshinterval * 2 will be the timeout for network endpoint
    '''
    network_list = list(network_ip_dict.keys())
    vxlanendpoint_list = [VXLANEndpointSet.default_key(n) for n in network_list]
    all_tun_ports2 = list(set(created_ports.keys()).union(set(removed_ports.keys())))
    def update_vxlanendpoints(keys, values, timestamp):
        # values = List[VXLANEndpointSet]
        # endpointlist is [src_ip, vhost, systemid, bridge, expire]
        for v,n in zip(values[0:len(network_list)], network_list):
            if v is not None:
                v.endpointlist = [ep for ep in v.endpointlist
                                  if (ep[1], ep[2], ep[3]) != (ovsdb_vhost, system_id, bridge)
                                  and ep[4] >= timestamp]
                ip_address = network_ip_dict[n]
                if ip_address is not None:
                    v.endpointlist.append([ip_address,
                              ovsdb_vhost,
                              system_id,
                              bridge,
                              None if refreshinterval is None else
                                    refreshinterval * 1000000 * 2 + timestamp
                              ])
        written_values = {}
        if all_tun_ports2:
            for k,v,vxkey,vxinfo in zip(keys[len(network_list):len(network_list) + len(all_tun_ports2)],
                           values[len(network_list):len(network_list) + len(all_tun_ports2)],
                           keys[len(network_list) + len(all_tun_ports2):len(network_list) + 2 * len(all_tun_ports2)],
                           values[len(network_list) + len(all_tun_ports2):len(network_list) + 2 * len(all_tun_ports2)]):
                if v is None:
                    if vxinfo is not None:
                        # The port is deleted? Then we should also delete the vxinfo
                        written_values[vxkey] = None
                else:
                    if v.id in created_ports:
                        if vxinfo is None:
                            vxinfo = LogicalPortVXLANInfo.create_from_key(vxkey)
                        # There maybe more than one endpoint at the same time (on migrating)
                        # so we keep all possible endpoints, but move our endpoint to the first place
                        myendpoint = {'vhost': ovsdb_vhost,
                                      'systemid': system_id,
                                      'bridge': bridge,
                                      'tunnel_dst': created_ports[v.id],
                                      'updated_time': timestamp}
                        vxinfo.endpoints = [ep for ep in vxinfo.endpoints
                                            if ep['updated_time'] + allowedmigrationtime * 1000000 >= timestamp
                                            and (ep['vhost'], ep['systemid'], ep['bridge']) != (ovsdb_vhost, system_id, bridge)]
                        vxinfo.endpoints = [myendpoint] + vxinfo.endpoints
                        written_values[vxkey] = vxinfo
                    elif v.id in removed_ports:
                        if vxinfo is not None:
                            # Remove endpoint
                            vxinfo.endpoints = [ep for ep in vxinfo.endpoints
                                                if (ep['vhost'], ep['systemid'], ep['bridge']) != (ovsdb_vhost, system_id, bridge)]
                            if not vxinfo.endpoints:
                                written_values[vxkey] = None
                            else:
                                written_values[vxkey] = vxinfo
        written_values_list = tuple(written_values.items())
        return (tuple(itertools.chain(keys[:len(network_list)], (k for k,_ in written_values_list))),
                tuple(itertools.chain(values[:len(network_list)], (v for _,v in written_values_list))))
    for m in callAPI(container, 'objectdb', 'transact', {'keys': tuple(vxlanendpoint_list + [LogicalPort.default_key(p) for p in all_tun_ports2] +
                                                                  [LogicalPortVXLANInfo.default_key(p) for p in all_tun_ports2]),
                                                    'updater': update_vxlanendpoints,
                                                    'withtime': True
                                                    }):
        yield m

def _get_ip(ip):
    try:
        return ip4_addr(ip)
    except Exception:
        return None

def get_broadcast_ips(vxlan_endpointset, local_ip, ovsdb_vhost, system_id, bridge):
    '''
    Get all IP addresses that are not local
    
    :param vxlan_endpointset: a VXLANEndpointSet object
    
    :param local_ips: list of local IP address to exclude with
    
    :param ovsdb_vhost: identifier, vhost
    
    :param system_id: identifier, system-id
    
    :param bridge: identifier, bridge name
    
    :return: `[(ip, ipnum)]` list where IPs are the original string of the IP address, and ipnum
             are 32-bit numeric IPv4 address.
    '''
    localip_addr = _get_ip(local_ip)
    allips = [(ip, ipnum) for ip, ipnum in ((ep[0], _get_ip(ep[0])) for ep in vxlan_endpointset.endpointlist
              if (ep[1], ep[2], ep[3]) != (ovsdb_vhost, system_id, bridge))
              if ipnum is not None and ipnum != localip_addr]
    return allips



'''
Created on 2015/06/01

:author: hubo
'''
from __future__ import print_function, absolute_import, division 
import copy
import warnings

try:
    from vlcp_event_cython.event import *
except:

    class IsMatchExceptionWarning(Warning):
        pass

    class EventMatcher(object):
        '''
        A matcher to match an event
        '''
        def __init__(self, indices, judgeFunc = None):
            # cut indices into real size
            for i in range(len(indices) - 1, -1, -1):
                if indices[i] is not None:
                    break
            self.indices = indices[:i+1]
            if judgeFunc is not None:
                def _warning_judge(e):
                    try:
                        return judgeFunc(e)
                    except Exception as exc:
                        # Do not crash
                        warnings.warn(IsMatchExceptionWarning('Exception raised when _ismatch is calculated: %r. event = %r, matcher = %r, _ismatch = %r'
                                                              % (exc, e, self, judgeFunc)))
                        return False
                self.judge = _warning_judge
        def judge(self, event):
            return True
        def isMatch(self, event, indexStart = 0):
            if len(self.indices) > len(event.indices):
                return False
            for i in range(indexStart, len(self.indices)):
                if self.indices[i] is not None and self.indices[i] != event.indices[i]:
                    return False
            return self.judge(event)
        def __repr__(self):
            cls = type(self)
            return '<EventMatcher:' + \
                repr(self.indices) + '>'
        
    def withIndices(*args):
        '''
        Create indices for an event class. Every event class must be decorated with this decorator.
        '''
        def decorator(cls):
            for c in cls.__bases__:
                if hasattr(c, '_indicesNames'):
                    cls._classnameIndex = c._classnameIndex + 1
                    for i in range(0, cls._classnameIndex):
                        setattr(cls, '_classname' + str(i), getattr(c, '_classname' + str(i)))
                    setattr(cls, '_classname' + str(cls._classnameIndex), cls._getTypename())
                    cls._indicesNames = c._indicesNames + ('_classname' + str(cls._classnameIndex),) + args
                    cls._generateTemplate()
                    return cls
            cls._classnameIndex = -1
            cls._indicesNames = args
            cls._generateTemplate()
            return cls
        return decorator

    class Event(object):
        '''
        A generated event with indices
        '''
        canignore = True
        _indicesNames = ()
        _classnameIndex = -1
        def __init__(self, *args, **kwargs):
            '''
            :param args: index values like 12,"read",... content are type-depended.
            
            :param kwargs:
            
                *indices*
                    input indices by name
                    
                canignore
                    if the event is not processed, whether it is safe to ignore the event.
                    
                    If it is not, the processing queue might be blocked to wait for a proper event processor.
                    Default to True.
                    
                *others*
                    the properties will be set on the created event
            '''
            indicesNames = self.indicesNames()
            if kwargs and not args:
                indices = tuple(kwargs[k] if k[:10] != '_classname' else getattr(self, k) for k in indicesNames)
            else:
                indices = tuple(self._generateIndices(args))
            self.indices = indices
            for k, v in zip(indicesNames, indices):
                setattr(self, k, v)
            for k, v in kwargs.items():
                if k not in indicesNames:
                    setattr(self, k, v)
        @classmethod
        def indicesNames(cls):
            '''
            :returns: names of indices
            '''
            return getattr(cls, '_indicesNames', ())
        @classmethod
        def _getTypename(cls):
            module = cls.__module__
            if module is None:
                return cls.__name__
            else:
                return module  + '.' + cls.__name__        
        @classmethod
        def getTypename(cls):
            '''
            :returns: return the proper name to match
            '''
            if cls is Event:
                return None
            else:
                for c in cls.__bases__:
                    if issubclass(c, Event):
                        if c is Event:
                            return cls._getTypename()
                        else:
                            return c.getTypename()
        @classmethod
        def _generateTemplate(cls):
            names = cls.indicesNames()
            template = [None] * len(names)
            argpos = []
            leastsize = 0
            for i in range(0, len(names)):
                if names[i][:10] == '_classname':
                    template[i] = getattr(cls, names[i])
                    leastsize = i + 1
                else:
                    argpos.append(i)
            cls._template = template
            cls._argpos = argpos
            cls._leastsize = leastsize
        @classmethod
        def _generateIndices(cls, args):
            indices = cls._template[:]
            ap = cls._argpos
            lp = 0
            if args:
                for i in range(0, len(args)):
                    indices[ap[i]] = args[i]
                lp = ap[len(args) - 1] + 1
            return indices[:max(cls._leastsize, lp)]
        @classmethod
        def createMatcher(cls, *args, **kwargs):
            '''
            :param _ismatch: user-defined function ismatch(event) for matching test
            :param \*args: indices
            :param \*\*kwargs: index_name=index_value for matching criteria
            '''
            if kwargs and not args:
                return EventMatcher(tuple(getattr(cls, ind) if ind[:10] == '_classname' else kwargs.get(ind) for ind in cls.indicesNames()), kwargs.get('_ismatch'))
            else:
                return EventMatcher(tuple(cls._generateIndices(args)), kwargs.get('_ismatch'))
        def __repr__(self):
            cls = type(self)
            return '<' + cls.__module__ + '.' + cls.__name__ + '(' + self.getTypename() + '): {' + \
                ', '.join(repr(k) + ': ' + repr(v) for k,v in zip(self.indicesNames(), self.indices)) + '}>'
        def canignorenow(self):
            '''
            Extra criteria for an event with canignore = False.
            When this event returns True, the event is safely ignored.
            '''
            return False




'''
Created on 2015/10/19

:author: hubo

Remove pycache files from a module, to ensure a successful reload
'''
import os

try:
    reload
except:
    try:
        from importlib import reload
    except:
        from imp import reload

def removeCache(module):
    if hasattr(module, '__cached__'):
        try:
            os.remove(module.__cached__)
        except:
            pass
    else:
        f = module.__file__
        try:
            if f.endswith('.pyc'):
                os.remove(f)
            elif f.endswith('.py'):
                os.remove(f + 'c')
        except:
            pass




import json

from vlcp.config import defaultconfig

from vlcp.event import RoutineContainer

from vlcp.utils.webclient import WebClient, WebException

from vlcp.server.module import Module,callAPI, api

@defaultconfig
class RemoteCall(Module):
    """
    Route local API calls to remote management API.
    """
    # URL list for every target, should be ``{target: [url, url, ...]}``
    _default_target_url_map = {}

    def __init__(self,server):
        super(RemoteCall,self).__init__(server)
        self.app_routine = RoutineContainer(self.scheduler)
        # there is no need to run this container main, so don't append it
        # self.routines.append(self.app_routine)
        self.wc = WebClient()
        self.app_routine.main = self._main
        self.routines.append(self.app_routine)
        self.createAPI(api(self.call,self.app_routine))
    
    def _main(self):
        try:
            self.wc.cleanup_task(self.app_routine, 120)
            yield ()
        finally:
            self.wc.endtask()

    def call(self,remote_module,method,timeout,params):
        """
        Call remote API
        
        :param remote_module: target name for the remote module
        
        :param method: method name of the API
        
        :param timeout: timeout for the call
        
        :param params: A dictionary contains all the parameters need for the call
        
        :return: Return result from the remote call
        """
        self._logger.info("remote call remote_module %r",remote_module)
        self._logger.info("remote call method %r", method)
        self._logger.info("remote call kwargs %r", params)
        success = False

        if remote_module and remote_module in self.target_url_map:
            endpoints = self.target_url_map[remote_module]
            params = json.dumps(params).encode("utf-8")
            for endpoint in endpoints:
                url = endpoint + "/" + remote_module + "/" + method

                try:
                    for m in self.wc.urlgetcontent(self.app_routine,url,params,b'POST',
                                                   {"Content-Type":"application/json"},timeout=timeout):
                        yield m
                except WebException as e:
                    # this endpoint connection error , try next url
                    self._logger.warning(" url (%r) post error %r , break ..",url,e)
                    success = False
                    raise
                except Exception:
                    self._logger.warning(" url (%r) connection error , try next ..", url)
                    continue
                else:
                    success = True
                    break

        else:
            self._logger.warning(" target (%r) url not existed, ignore it ",remote_module)

        if not success:
            raise IOError("remote call connection error !")

        self.app_routine.retvalue = []

def remoteAPI(container,targetname,name,params={},timeout=60.0):
    args = {"remote_module":targetname,"method":name,"timeout":timeout,"params":params}
    for m in callAPI(container,"remotecall","call",params=args,timeout=timeout + 20):
        yield m


'''
Created on 2016/5/24

:author: hubo
'''


from vlcp.config.config import defaultconfig
from vlcp.server.module import depend
from vlcp.service.sdn.flowbase import FlowBase
from vlcp.service.sdn.ofpmanager import FlowInitialize
from vlcp.utils.flowupdater import FlowUpdater
from vlcp.utils.networkmodel import PhysicalPort, LogicalPort, LogicalNetwork
import vlcp.service.kvdb.objectdb as objectdb
import vlcp.service.sdn.ofpportmanager as ofpportmanager
from vlcp.event.runnable import RoutineContainer
from vlcp.protocol.openflow.openflow import OpenflowConnectionStateEvent,\
    OpenflowAsyncMessageEvent, OpenflowErrorResultException
from vlcp.utils.ethernet import ethernet_l2
import vlcp.service.sdn.ioprocessing as iop
import itertools

class L2FlowUpdater(FlowUpdater):

    def __init__(self, connection, parent):
        FlowUpdater.__init__(self, connection, (), ('l2switch', connection), parent._logger)
        self._parent = parent
        self._lastlognets = ()
        self._lastlogports = ()
        self._lastphyports = ()
        self._lastlogportinfo = {}
        self._lastphyportinfo = {}
        self._lastlognetinfo = {}
    def main(self):
        try:
            self.subroutine(self._update_handler_prepush(), name = '_update_handler_routine')
            for m in FlowUpdater.main(self):
                yield m
        finally:
            self._update_handler_routine.close()
    def _walk_logport(self, key, value, walk, save):
        save(key)
        if value is None:
            return
        try:
            net = walk(value.network.getkey())
        except KeyError:
            pass
        else:
            save(net.getkey())
    def _walk_phyport(self, key, value, walk, save):
        save(key)
    def _update_handler_prepush(self):
        dataobjectchanged = iop.DataObjectChanged.createMatcher(None, None, self._connection)
        while True:
            yield (dataobjectchanged,)
            self._lastlogports, self._lastphyports, self._lastlognets, _ = self.event.current
            self._initialkeys = [p.getkey() for p,_ in self._lastlogports] + \
                                [p.getkey() for p,_ in self._lastphyports]
            self._walkerdict = dict(itertools.chain(((p.getkey(), self._walk_logport) for p,_ in self._lastlogports),
                                                    ((p.getkey(), self._walk_phyport) for p,_ in self._lastphyports)))
            self.subroutine(self.restart_walk(), False)
    def updateflow(self, conn, addvalues, removevalues, updatedvalues):
        ofdef = conn.openflowdef
        vhost = conn.protocol.vhost
        l2out = self._parent._gettableindex('l2output', vhost)
        l2out_next = self._parent._getnexttable('', 'l2output', vhost = vhost)
        def _create_flows(networkid, macaddr, portid):
            if conn.protocol.disablenxext:
                # Use METADATA
                masked_network = ((networkid & 0xffff) << 32)
                return (ofdef.ofp_flow_mod(table_id = l2out,
                                            cookie = 0x3,
                                            cookie_mask = 0xffffffffffffffff,
                                            command = ofdef.OFPFC_ADD,
                                            priority = ofdef.OFP_DEFAULT_PRIORITY + 2,
                                            buffer_id = ofdef.OFP_NO_BUFFER,
                                           out_port = ofdef.OFPP_ANY,
                                           out_group = ofdef.OFPG_ANY,
                                            match = ofdef.ofp_match_oxm(
                                                        oxm_fields = [
                                                            ofdef.create_oxm(ofdef.OXM_OF_METADATA_W,
                                                                             masked_network,
                                                                             b'\x00\x00\xff\xff\x00\x00\x00\x00'),
                                                            ofdef.create_oxm(ofdef.OXM_OF_ETH_DST, macaddr)
                                                            ]
                                                    ),
                                            instructions = [ofdef.ofp_instruction_write_metadata(metadata = (portid & 0xffff),
                                                                                                 metadata_mask = 0xffff),
                                                            ofdef.ofp_instruction_goto_table(table_id = l2out_next)]
                                    ),)
            else:
                # Use REG5, REG6
                out_net = networkid
                return (ofdef.ofp_flow_mod(table_id = l2out,
                                        cookie = 0x3,
                                        cookie_mask = 0xffffffffffffffff,
                                        command = ofdef.OFPFC_ADD,
                                        priority = ofdef.OFP_DEFAULT_PRIORITY + 2,
                                        buffer_id = ofdef.OFP_NO_BUFFER,
                                       out_port = ofdef.OFPP_ANY,
                                       out_group = ofdef.OFPG_ANY,
                                        match = ofdef.ofp_match_oxm(
                                                    oxm_fields = [
                                                        ofdef.create_oxm(ofdef.NXM_NX_REG5, out_net),
                                                        ofdef.create_oxm(ofdef.OXM_OF_ETH_DST, macaddr)
                                                        ]
                                                ),
                                        instructions = [ofdef.ofp_instruction_actions(
                                                                actions = [ofdef.ofp_action_set_field(
                                                                                field = ofdef.create_oxm(ofdef.NXM_NX_REG6, portid)
                                                                            )],
                                                                type = ofdef.OFPIT_APPLY_ACTIONS
                                                                                      ),
                                                        ofdef.ofp_instruction_goto_table(table_id = l2out_next)]
                                ),)
        def _delete_flows(networkid, macaddr):
            if conn.protocol.disablenxext:
                # Use METADATA
                masked_network = ((networkid & 0xffff) << 32)
                return (ofdef.ofp_flow_mod(table_id = l2out,
                                            cookie = 0x3,
                                            cookie_mask = 0xffffffffffffffff,
                                            command = ofdef.OFPFC_DELETE_STRICT,
                                            priority = ofdef.OFP_DEFAULT_PRIORITY + 2,
                                            buffer_id = ofdef.OFP_NO_BUFFER,
                                            out_port = ofdef.OFPP_ANY,
                                            out_group = ofdef.OFPG_ANY,
                                            match = ofdef.ofp_match_oxm(
                                                        oxm_fields = [
                                                            ofdef.create_oxm(ofdef.OXM_OF_METADATA_W,
                                                                             masked_network,
                                                                             b'\x00\x00\xff\xff\x00\x00\x00\x00'),
                                                            ofdef.create_oxm(ofdef.OXM_OF_ETH_DST, macaddr)
                                                            ]
                                                    )
                                    ),)
            else:
                # Use REG5, REG6
                out_net = networkid
                return (ofdef.ofp_flow_mod(table_id = l2out,
                                            cookie = 0x3,
                                            cookie_mask = 0xffffffffffffffff,
                                            command = ofdef.OFPFC_DELETE_STRICT,
                                            priority = ofdef.OFP_DEFAULT_PRIORITY + 2,
                                            buffer_id = ofdef.OFP_NO_BUFFER,
                                            out_port = ofdef.OFPP_ANY,
                                            out_group = ofdef.OFPG_ANY,
                                            match = ofdef.ofp_match_oxm(
                                                        oxm_fields = [
                                                            ofdef.create_oxm(ofdef.NXM_NX_REG5, out_net),
                                                            ofdef.create_oxm(ofdef.OXM_OF_ETH_DST, macaddr)
                                                            ]
                                                    )
                                    ),)
        def _create_default_flow(networkid, portid):
            if conn.protocol.disablenxext:
                # Use METADATA
                masked_network = ((networkid & 0xffff) << 32)
                return (ofdef.ofp_flow_mod(table_id = l2out,
                                            cookie = 0x4,
                                            cookie_mask = 0xffffffffffffffff,
                                            command = ofdef.OFPFC_ADD,
                                            priority = ofdef.OFP_DEFAULT_PRIORITY + 1,
                                            buffer_id = ofdef.OFP_NO_BUFFER,
                                            out_port = ofdef.OFPP_ANY,
                                            out_group = ofdef.OFPG_ANY,
                                            match = ofdef.ofp_match_oxm(
                                                        oxm_fields = [
                                                            ofdef.create_oxm(ofdef.OXM_OF_METADATA_W,
                                                                             masked_network,
                                                                             b'\x00\x00\xff\xff\x00\x00\x00\x00')
                                                            ]
                                                    ),
                                            instructions = [ofdef.ofp_instruction_write_metadata(metadata = (portid & 0xffff),
                                                                                                 metadata_mask = 0xffff),
                                                            ofdef.ofp_instruction_goto_table(table_id = l2out_next)]
                                    ),)
            else:
                # Use REG5, REG6
                out_net = networkid
                return (ofdef.ofp_flow_mod(table_id = l2out,
                                            cookie = 0x4,
                                            cookie_mask = 0xffffffffffffffff,
                                            command = ofdef.OFPFC_ADD,
                                            priority = ofdef.OFP_DEFAULT_PRIORITY + 1,
                                            buffer_id = ofdef.OFP_NO_BUFFER,
                                            out_port = ofdef.OFPP_ANY,
                                            out_group = ofdef.OFPG_ANY,
                                            match = ofdef.ofp_match_oxm(
                                                        oxm_fields = [
                                                            ofdef.create_oxm(ofdef.NXM_NX_REG5, out_net)
                                                            ]
                                                    ),
                                            instructions = [ofdef.ofp_instruction_actions(
                                                                    actions = [ofdef.ofp_action_set_field(
                                                                                    field = ofdef.create_oxm(ofdef.NXM_NX_REG6, portid)
                                                                                )],
                                                                    type = ofdef.OFPIT_APPLY_ACTIONS
                                                                                          ),
                                                            ofdef.ofp_instruction_goto_table(table_id = l2out_next)]
                                    ),)
        def _delete_default_flows(networkid):
            if conn.protocol.disablenxext:
                # Use METADATA
                masked_network = ((networkid & 0xffff) << 32)
                return (ofdef.ofp_flow_mod(table_id = l2out,
                                            cookie = 0x4,
                                            cookie_mask = 0xffffffffffffffff,
                                            command = ofdef.OFPFC_DELETE_STRICT,
                                            priority = ofdef.OFP_DEFAULT_PRIORITY + 1,
                                            buffer_id = ofdef.OFP_NO_BUFFER,
                                            out_port = ofdef.OFPP_ANY,
                                            out_group = ofdef.OFPG_ANY,
                                            match = ofdef.ofp_match_oxm(
                                                        oxm_fields = [
                                                            ofdef.create_oxm(ofdef.OXM_OF_METADATA_W,
                                                                             masked_network,
                                                                             b'\x00\x00\xff\xff\x00\x00\x00\x00')
                                                            ]
                                                    )
                                    ),)
            else:
                # Use REG5, REG6
                out_net = networkid
                return (ofdef.ofp_flow_mod(table_id = l2out,
                                            cookie = 0x4,
                                            cookie_mask = 0xffffffffffffffff,
                                            command = ofdef.OFPFC_DELETE_STRICT,
                                            priority = ofdef.OFP_DEFAULT_PRIORITY + 1,
                                            buffer_id = ofdef.OFP_NO_BUFFER,
                                            out_port = ofdef.OFPP_ANY,
                                            out_group = ofdef.OFPG_ANY,
                                            match = ofdef.ofp_match_oxm(
                                                        oxm_fields = [
                                                            ofdef.create_oxm(ofdef.NXM_NX_REG5, out_net)
                                                            ]
                                                    )
                                    ),)
        try:

            lastlognetinfo = self._lastlognetinfo
            lastlogportinfo = self._lastlogportinfo
            allresult = set(v for v in self._savedresult if v is not None and not v.isdeleted())

            # Select one physical port for each physical network, remove others
            currentphynets = dict(sorted(((p.physicalnetwork, p) for p, _ in self._lastphyports if p in allresult),
                                         key=lambda x: x[1].getkey()))

            currentphyportinfo = dict((p,(p.physicalnetwork, pid)) for p,pid in self._lastphyports
                                      if p in allresult and p.physicalnetwork in currentphynets
                                      and currentphynets[p.physicalnetwork] == p)

            phynet_to_phyport_id_info = dict((p.physicalnetwork,v[1]) for p,v in currentphyportinfo.items())
            currentlognetinfo = dict((n, (nid, phynet_to_phyport_id_info.get(n.physicalnetwork,None)))
                                     for n, nid in self._lastlognets
                                     if n in allresult)

            def _try_create_macaddress(port):
                mac_address = getattr(port, 'mac_address', None)
                if mac_address is None:
                    return None
                else:
                    try:
                        return ofdef.mac_addr(mac_address)
                    except Exception:
                        return None
            currentlogportinfo = dict((logport, (_try_create_macaddress(logport), logportid,
                                                 currentlognetinfo.get(logport.network)[0]))
                                      for logport, logportid in self._lastlogports
                                      if logport in allresult
                                      and logport.network in currentlognetinfo)

            self._lastlognetinfo = currentlognetinfo
            self._lastlogportinfo = currentlogportinfo

            cmds = []
            for lognet in lastlognetinfo.keys():
                if lognet not in currentlognetinfo:
                    # this lognet has been removed
                    network_id , _ = lastlognetinfo[lognet]
                    cmds.extend(_delete_default_flows(network_id))
                elif lognet in currentlognetinfo and currentlognetinfo[lognet] != lastlognetinfo[lognet]:
                    # some info has been changed , remove old
                    network_id, _ = lastlognetinfo[lognet]
                    cmds.extend(_delete_default_flows(network_id))


            for logport in lastlogportinfo.keys():
                if logport not in currentlogportinfo:
                    # this logport has been removed
                    mac, _, network_id = lastlogportinfo[logport]

                    if mac:
                        cmds.extend(_delete_flows(network_id, mac))
                elif logport in currentlogportinfo and currentlogportinfo[logport] != lastlogportinfo[logport]:
                    # some info has been changed , remove old
                    mac, _ , network_id = lastlogportinfo[logport]

                    if mac:
                        cmds.extend(_delete_flows(network_id, mac))

            for m in self.execute_commands(conn, cmds):
                yield m

            del cmds[:]

            for lognet in currentlognetinfo.keys():
                if lognet not in lastlognetinfo:
                    # this lognet is new add
                    network_id, phyport_id = currentlognetinfo[lognet]
                    if phyport_id:
                        cmds.extend(_create_default_flow(network_id,phyport_id))
                elif lognet in lastlognetinfo and lastlognetinfo[lognet] != currentlognetinfo[lognet]:
                    # this lognet info changed , add new info
                    network_id, phyport_id = currentlognetinfo[lognet]

                    if phyport_id:
                        cmds.extend(_create_default_flow(network_id, phyport_id))

            for logport in currentlogportinfo.keys():
                if logport not in lastlogportinfo:
                    # this logport is new add
                    mac, port_id, network_id = currentlogportinfo[logport]
                    if mac:
                        cmds.extend(_create_flows(network_id, mac,port_id))
                elif logport in lastlogportinfo and lastlogportinfo[logport] != currentlogportinfo[logport]:
                    # this logport info changed, add new info
                    mac, port_id, network_id = currentlogportinfo[logport]
                    if mac:
                        cmds.extend(_create_flows(network_id, mac, port_id))

            for m in self.execute_commands(conn,cmds):
                yield m

        except Exception:
            self._parent._logger.warning("Update l2switch flow for connection %r failed with exception", conn, exc_info = True)
            # We don't want the whole flow update stops, so ignore the exception and continue
            
    

@defaultconfig
@depend(ofpportmanager.OpenflowPortManager, objectdb.ObjectDB)
class L2Switch(FlowBase):
    "L2 switch functions"
    _tablerequest = (("l2input", ('ingress',), ''),
                     ("l2output", ('l2input',), ''),
                     ('egress', ('l2output', 'l2learning'), ''),
                     ("l2learning", ('l2output',), 'l2learning'))
    # Use learning or prepush. Learning strategy creates flows on necessary, but may cause
    # some delay for the first packet; prepush push all necessary flows from beginning.
    _default_learning = False
    # When using learning=True, enable OpenvSwitch learn() action. if nxlearn=False, the
    # learning strategy will be done with controller PACKET_IN processing, which may be
    # less efficient.
    _default_nxlearn = True
    # Timeout for a learned flow
    _default_learntimeout = 300
    def __init__(self, server):
        FlowBase.__init__(self, server)
        self.apiroutine = RoutineContainer(self.scheduler)
        self.apiroutine.main = self._main
        self.routines.append(self.apiroutine)
        self._flowupdaters = {}
    def _main(self):
        flow_init = FlowInitialize.createMatcher(_ismatch = lambda x: self.vhostbind is None or x.vhost in self.vhostbind)
        conn_down = OpenflowConnectionStateEvent.createMatcher(state = OpenflowConnectionStateEvent.CONNECTION_DOWN,
                                                               _ismatch = lambda x: self.vhostbind is None or x.createby.vhost in self.vhostbind)
        while True:
            yield (flow_init, conn_down)
            if self.apiroutine.matcher is flow_init:
                c = self.apiroutine.event.connection
                self.apiroutine.subroutine(self._init_conn(self.apiroutine.event.connection))
            else:
                c = self.apiroutine.event.connection
                self.apiroutine.subroutine(self._remove_conn(c))
    def _init_conn(self, conn):
        # Default
        ofdef = conn.openflowdef
        vhost = conn.protocol.vhost
        l2 = self._gettableindex('l2input', vhost)
        l2_next = self._getnexttable('', 'l2input', vhost = vhost)
        l2out = self._gettableindex('l2output', vhost)
        l2out_next = self._getnexttable('', 'l2output', vhost = vhost)
        l2learning = self._gettableindex('l2learning', vhost)
        if hasattr(conn, '_l2switch_learning_routine') and conn._l2switch_learning_routine:
            conn._l2switch_learning_routine.close()
            delattr(conn, '_l2switch_learning_routine')
        if conn in self._flowupdaters:
            self._flowupdaters[conn].close()
            del self._flowupdaters[conn]
        if not self.learning:
            new_updater = L2FlowUpdater(conn, self)
            self._flowupdaters[conn] = new_updater
            new_updater.start()
        if self.learning:
            if self.nxlearn and not conn.protocol.disablenxext:
                # Use nx_action_learn
                for m in conn.protocol.batch((ofdef.ofp_flow_mod(table_id = l2,
                                                                   command = ofdef.OFPFC_ADD,
                                                                   priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                                   match = ofdef.ofp_match_oxm(
                                                                                oxm_fields = [
                                                                                    # Drop packets with an broadcast MAC address as dl_src
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_SRC_W, b'\x01\x00\x00\x00\x00\x00', b'\x01\x00\x00\x00\x00\x00')
                                                                                    ]
                                                                            ),
                                                                   out_port = ofdef.OFPP_ANY,
                                                                   out_group = ofdef.OFPG_ANY,
                                                                   instructions = [ofdef.ofp_instruction_actions(
                                                                                    type = ofdef.OFPIT_CLEAR_ACTIONS
                                                                                    )]
                                                                   ),
                                            ofdef.ofp_flow_mod(table_id = l2,
                                                                   command = ofdef.OFPFC_ADD,
                                                                   priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                                   match = ofdef.ofp_match_oxm(
                                                                                oxm_fields = [
                                                                                    # Drop 802.1D STP packets, they should never be forwarded
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_DST_W, b'\x01\x80\xc2\x00\x00\x00', b'\xff\xff\xff\xff\xff\xf0')
                                                                                    ]
                                                                            ),
                                                                   out_port = ofdef.OFPP_ANY,
                                                                   out_group = ofdef.OFPG_ANY,
                                                                   instructions = [ofdef.ofp_instruction_actions(
                                                                                    type = ofdef.OFPIT_CLEAR_ACTIONS
                                                                                    )]
                                                                   ),
                                            ofdef.ofp_flow_mod(table_id = l2,
                                                                cookie = 0x1,
                                                                cookie_mask = 0xffffffffffffffff,
                                                                   command = ofdef.OFPFC_ADD,
                                                                   priority = 0,
                                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                                   match = ofdef.ofp_match_oxm(),
                                                                   out_port = ofdef.OFPP_ANY,
                                                                   out_group = ofdef.OFPG_ANY,
                                                                   instructions = [ofdef.ofp_instruction_actions(
                                                                                    actions = [
                                                                                        ofdef.nx_action_learn(
                                                                                                hard_timeout = self.learntimeout,
                                                                                                priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                                                                cookie = 0x2,
                                                                                                table_id = l2learning,
                                                                                                specs = [ofdef.create_nxfms_matchfield(ofdef.NXM_NX_REG4, ofdef.NXM_NX_REG5),
                                                                                                         ofdef.create_nxfms_matchfield(ofdef.NXM_OF_ETH_SRC, ofdef.NXM_OF_ETH_DST),
                                                                                                         ofdef.create_nxfms_loadfield(ofdef.OXM_OF_IN_PORT, ofdef.NXM_NX_REG6)]
                                                                                                )
                                                                                               ],
                                                                                    type = ofdef.OFPIT_APPLY_ACTIONS
                                                                                    ),
                                                                                   ofdef.ofp_instruction_goto_table(table_id = l2_next)]
                                                                   ),
                                              ofdef.ofp_flow_mod(table_id = l2out,
                                                                   command = ofdef.OFPFC_ADD,
                                                                   priority = ofdef.OFP_DEFAULT_PRIORITY + 10,
                                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                                   out_port = ofdef.OFPP_ANY,
                                                                   out_group = ofdef.OFPG_ANY,
                                                                   match = ofdef.ofp_match_oxm(
                                                                                oxm_fields = [
                                                                                    # A broadcast packet
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_DST_W, b'\x01\x00\x00\x00\x00\x00', b'\x01\x00\x00\x00\x00\x00')
                                                                                    ]
                                                                            ),
                                                                   instructions = [ofdef.ofp_instruction_goto_table(table_id = l2out_next)]
                                                                   ),
                                              ofdef.ofp_flow_mod(table_id = l2out,
                                                                cookie = 0x1,
                                                                cookie_mask = 0xffffffffffffffff,
                                                                   command = ofdef.OFPFC_ADD,
                                                                   priority = 0,
                                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                                   out_port = ofdef.OFPP_ANY,
                                                                   out_group = ofdef.OFPG_ANY,
                                                                   match = ofdef.ofp_match_oxm(),
                                                                   instructions = [ofdef.ofp_instruction_actions(
                                                                                    actions = [
                                                                                        ofdef.nx_action_resubmit(
                                                                                                table = l2learning,
                                                                                                in_port = ofdef.nx_port_no.OFPP_IN_PORT
                                                                                                )
                                                                                               ],
                                                                                    type = ofdef.OFPIT_APPLY_ACTIONS
                                                                                    ),
                                                                                   ofdef.ofp_instruction_goto_table(table_id = l2out_next)]
                                                                   )), conn, self.apiroutine):
                    yield m
            else:
                # Use PACKET_IN
                for m in conn.protocol.batch((ofdef.ofp_flow_mod(table_id = l2,
                                                                   command = ofdef.OFPFC_ADD,
                                                                   priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                                   out_port = ofdef.OFPP_ANY,
                                                                   out_group = ofdef.OFPG_ANY,
                                                                   match = ofdef.ofp_match_oxm(
                                                                                oxm_fields = [
                                                                                    # Drop packets with an broadcast MAC address as dl_src
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_SRC_W, b'\x01\x00\x00\x00\x00\x00', b'\x01\x00\x00\x00\x00\x00')
                                                                                    ]
                                                                            ),
                                                                   instructions = [ofdef.ofp_instruction_actions(
                                                                                    type = ofdef.OFPIT_CLEAR_ACTIONS
                                                                                    )]
                                                                   ),
                                            ofdef.ofp_flow_mod(table_id = l2,
                                                                   command = ofdef.OFPFC_ADD,
                                                                   priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                                   out_port = ofdef.OFPP_ANY,
                                                                   out_group = ofdef.OFPG_ANY,
                                                                   match = ofdef.ofp_match_oxm(
                                                                                oxm_fields = [
                                                                                    # Drop 802.1D STP packets, they should never be forwarded
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_DST_W, b'\x01\x80\xc2\x00\x00\x00', b'\xff\xff\xff\xff\xff\xf0')
                                                                                    ]
                                                                            ),
                                                                   instructions = [ofdef.ofp_instruction_actions(
                                                                                    type = ofdef.OFPIT_CLEAR_ACTIONS
                                                                                    )]
                                                                   ),
                                            ofdef.ofp_flow_mod(table_id = l2,
                                                                cookie = 0x1,
                                                                cookie_mask = 0xffffffffffffffff,
                                                                   command = ofdef.OFPFC_ADD,
                                                                   priority = 1,
                                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                                   out_port = ofdef.OFPP_ANY,
                                                                   out_group = ofdef.OFPG_ANY,
                                                                   match = ofdef.ofp_match_oxm(),
                                                                   instructions = [ofdef.ofp_instruction_actions(
                                                                                    actions = [
                                                                                        ofdef.ofp_action_output(port = ofdef.OFPP_CONTROLLER,
                                                                                                                max_len = 32
                                                                                                                )
                                                                                               ],
                                                                                    type = ofdef.OFPIT_APPLY_ACTIONS
                                                                                    ),
                                                                                   ofdef.ofp_instruction_goto_table(table_id = l2_next)]
                                                                   ),
                                              ofdef.ofp_flow_mod(table_id = l2out,
                                                                   command = ofdef.OFPFC_ADD,
                                                                   priority = ofdef.OFP_DEFAULT_PRIORITY + 10,
                                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                                   out_port = ofdef.OFPP_ANY,
                                                                   out_group = ofdef.OFPG_ANY,
                                                                   match = ofdef.ofp_match_oxm(
                                                                                oxm_fields = [
                                                                                    # A broadcast packet
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_DST_W, b'\x01\x00\x00\x00\x00\x00', b'\x01\x00\x00\x00\x00\x00')
                                                                                    ]
                                                                            ),
                                                                   instructions = [ofdef.ofp_instruction_goto_table(table_id = l2out_next)]
                                                                   )
                                              ), conn, self.apiroutine):
                    yield m                    
                def learning_packet_handler():
                    packetin = OpenflowAsyncMessageEvent.createMatcher(ofdef.OFPT_PACKET_IN, None, None, l2, 1, conn, conn.connmark)
                    conndown = conn.protocol.statematcher(conn)
                    while True:
                        yield (packetin, conndown)
                        if conn.matcher is conndown:
                            break
                        else:
                            msg = conn.event.message
                            try:
                                p = ethernet_l2.create(msg.data)
                            except Exception:
                                self._logger.warning('Invalid packet received: %r', conn.event.message.data, exc_info = True)
                            else:
                                dl_src = p.dl_src
                                in_port = ofdef.get_oxm(msg.match.oxm_fields, ofdef.OXM_OF_IN_PORT)
                                if conn.protocol.disablenxext:
                                    # Use METADATA
                                    metadata = ofdef.get_oxm(msg.match.oxm_fields, ofdef.OXM_OF_METADATA)
                                    masked_network = b'\x00\x00' + metadata[0:2] + b'\x00\x00\x00\x00'
                                    cmds = [ofdef.ofp_flow_mod(table_id = l2,
                                                                cookie = 0x2,
                                                                cookie_mask = 0xffffffffffffffff,
                                                                   command = ofdef.OFPFC_ADD,
                                                                   priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                                   out_port = ofdef.OFPP_ANY,
                                                                   out_group = ofdef.OFPG_ANY,
                                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                                   hard_timeout = max((self.learntimeout//4), 1),
                                                                   match = ofdef.ofp_match_oxm(
                                                                                oxm_fields = [
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_METADATA_W,
                                                                                                     metadata[0:2] + b'\x00\x00\x00\x00\x00\x00',
                                                                                                     b'\xff\xff\x00\x00\x00\x00\x00\x00'),
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_SRC, dl_src),
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_IN_PORT, in_port)
                                                                                    ]
                                                                            ),
                                                                   instructions = [ofdef.ofp_instruction_goto_table(table_id = l2_next)]
                                                                   ),
                                                                ofdef.ofp_flow_mod(table_id = l2out,
                                                                    cookie = 0x2,
                                                                    cookie_mask = 0xffffffffffffffff,
                                                                    command = ofdef.OFPFC_ADD,
                                                                    priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                                    buffer_id = ofdef.OFP_NO_BUFFER,
                                                                    hard_timeout = self.learntimeout,
                                                                    out_port = ofdef.OFPP_ANY,
                                                                    out_group = ofdef.OFPG_ANY,
                                                                    match = ofdef.ofp_match_oxm(
                                                                                oxm_fields = [
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_METADATA_W,
                                                                                                     masked_network,
                                                                                                     b'\x00\x00\xff\xff\x00\x00\x00\x00'),
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_DST, dl_src)
                                                                                    ]
                                                                            ),
                                                                    instructions = [ofdef.ofp_instruction_write_metadata(metadata = ofdef.uint16.create(in_port[2:4]),
                                                                                                                         metadata_mask = 0xffff),
                                                                                    ofdef.ofp_instruction_goto_table(table_id = l2out_next)]
                                                            )]
                                else:
                                    # Use REG5, REG6
                                    out_net = ofdef.get_oxm(msg.match.oxm_fields, ofdef.NXM_NX_REG4)
                                    cmds = [ofdef.ofp_flow_mod(table_id = l2,
                                                                cookie = 0x2,
                                                                cookie_mask = 0xffffffffffffffff,
                                                                   command = ofdef.OFPFC_ADD,
                                                                   priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                                   hard_timeout = max((self.learntimeout//4), 1),
                                                                   out_port = ofdef.OFPP_ANY,
                                                                   out_group = ofdef.OFPG_ANY,
                                                                   match = ofdef.ofp_match_oxm(
                                                                                oxm_fields = [
                                                                                    ofdef.create_oxm(ofdef.NXM_NX_REG4, out_net),
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_SRC, dl_src),
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_IN_PORT, in_port)
                                                                                    ]
                                                                            ),
                                                                   instructions = [ofdef.ofp_instruction_goto_table(table_id = l2_next)]
                                                                   ),
                                                                ofdef.ofp_flow_mod(table_id = l2out,
                                                                    cookie = 0x2,
                                                                    cookie_mask = 0xffffffffffffffff,
                                                                    command = ofdef.OFPFC_ADD,
                                                                    priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                                    buffer_id = ofdef.OFP_NO_BUFFER,
                                                                    hard_timeout = self.learntimeout,
                                                                    out_port = ofdef.OFPP_ANY,
                                                                    out_group = ofdef.OFPG_ANY,
                                                                    match = ofdef.ofp_match_oxm(
                                                                                oxm_fields = [
                                                                                    ofdef.create_oxm(ofdef.NXM_NX_REG5, out_net),
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_DST, dl_src)
                                                                                    ]
                                                                            ),
                                                                    instructions = [ofdef.ofp_instruction_actions(
                                                                                            actions = [ofdef.ofp_action_set_field(
                                                                                                            field = ofdef.create_oxm(ofdef.NXM_NX_REG6, in_port)
                                                                                                        )],
                                                                                            type = ofdef.OFPIT_APPLY_ACTIONS
                                                                                                                  ),
                                                                                    ofdef.ofp_instruction_goto_table(table_id = l2out_next)]
                                                            )]
                                if msg.buffer_id != ofdef.OFP_NO_BUFFER:
                                    # Send a packet out without actions to drop the packet and release the buffer
                                    cmds.append(ofdef.ofp_packet_out(buffer_id = msg.buffer_id,
                                                                     in_port = ofdef.OFPP_CONTROLLER,
                                                                     actions = []
                                                                     ))
                                conn.subroutine(conn.protocol.batch(cmds, conn, conn))
                conn.subroutine(learning_packet_handler(), name = '_l2switch_learning_routine')
        else:
            # Disable learning
            for m in conn.protocol.batch((ofdef.ofp_flow_mod(table_id = l2,
                                                                   command = ofdef.OFPFC_ADD,
                                                                   priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                                   out_port = ofdef.OFPP_ANY,
                                                                   out_group = ofdef.OFPG_ANY,
                                                                   match = ofdef.ofp_match_oxm(
                                                                                oxm_fields = [
                                                                                    # Drop packets with an broadcast MAC address as dl_src
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_SRC_W, b'\x01\x00\x00\x00\x00\x00', b'\x01\x00\x00\x00\x00\x00')
                                                                                    ]
                                                                            ),
                                                                   instructions = [ofdef.ofp_instruction_actions(
                                                                                    type = ofdef.OFPIT_CLEAR_ACTIONS
                                                                                    )]
                                                                   ),
                                          ofdef.ofp_flow_mod(table_id = l2,
                                                                   command = ofdef.OFPFC_ADD,
                                                                   priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                                   out_port = ofdef.OFPP_ANY,
                                                                   out_group = ofdef.OFPG_ANY,
                                                                   match = ofdef.ofp_match_oxm(
                                                                                oxm_fields = [
                                                                                    # Drop 802.1D STP packets, they should never be forwarded
                                                                                    ofdef.create_oxm(ofdef.OXM_OF_ETH_DST_W, b'\x01\x80\xc2\x00\x00\x00', b'\xff\xff\xff\xff\xff\xf0')
                                                                                    ]
                                                                            ),
                                                                   instructions = [ofdef.ofp_instruction_actions(
                                                                                    type = ofdef.OFPIT_CLEAR_ACTIONS
                                                                                    )]
                                                                   ),
                                          ofdef.ofp_flow_mod(table_id = l2,
                                                            cookie = 0x1,
                                                            cookie_mask = 0xffffffffffffffff,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               command = ofdef.OFPFC_ADD,
                                                               priority = 0,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               match = ofdef.ofp_match_oxm(),
                                                               instructions = [ofdef.ofp_instruction_goto_table(table_id = l2_next)]
                                                               ),
                                          ofdef.ofp_flow_mod(table_id = l2out,
                                                               command = ofdef.OFPFC_ADD,
                                                               priority = ofdef.OFP_DEFAULT_PRIORITY + 10,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               match = ofdef.ofp_match_oxm(
                                                                            oxm_fields = [
                                                                                # A broadcast packet
                                                                                ofdef.create_oxm(ofdef.OXM_OF_ETH_DST_W, b'\x01\x00\x00\x00\x00\x00', b'\x01\x00\x00\x00\x00\x00')
                                                                                ]
                                                                        ),
                                                               instructions = [ofdef.ofp_instruction_goto_table(table_id = l2out_next)]
                                                               ),
                                          ofdef.ofp_flow_mod(table_id = l2out,
                                                            cookie = 0x1,
                                                            cookie_mask = 0xffffffffffffffff,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               command = ofdef.OFPFC_ADD,
                                                               priority = 0,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               match = ofdef.ofp_match_oxm(),
                                                               instructions = [ofdef.ofp_instruction_goto_table(table_id = l2out_next)]
                                                               ),
                                          ), conn, self.apiroutine):
                yield m
    def _remove_conn(self, conn):
        # Do not need to modify flows
        if conn in self._flowupdaters:
            self._flowupdaters[conn].close()
            del self._flowupdaters[conn]
        if False:
            yield



'''
Created on 2016/1/8

:author: hubo
'''
from vlcp.config.config import Configurable, config
from vlcp.event.connection import Client
from vlcp.protocol.redis import Redis, RedisConnectionStateEvent, RedisSubscribeMessageEvent,\
    RedisReplyException
from contextlib import contextmanager, closing

def _str(b, encoding = 'ascii'):
    if isinstance(b, str):
        return b
    elif isinstance(b, bytes):
        return b.decode(encoding)
    else:
        return str(b)

def _conn(func):
    def f(self, container, *args, **kwargs):
        for m in self._get_default_connection(container):
            yield m
        for m in func(self, container, *args, **kwargs):
            yield m
    return f

class RedisConnectionDown(IOError):
    pass

class RedisConnectionRestarted(RedisConnectionDown):
    pass

@config('redisclient')
class RedisClientBase(Configurable):
    '''
    Connect to Redis server
    '''
    # Default connection URL for Redis client
    _default_url = 'tcp://localhost/'
    # Wait for the connection setup before raise an exception
    _default_timeout = 10
    # Select database
    _default_db = 0
    def __init__(self, conn = None, parent = None, protocol = None):
        Configurable.__init__(self)
        self._defaultconn = conn
        self._parent = parent
        self._lockconnmark = None
        if protocol:
            self._protocol = protocol
        else:
            if parent:
                self._protocol = parent._protocol
            else:
                self._protocol = Redis()
    def _get_connection(self, container, connection):
        if not connection.connected:
            for m in container.waitWithTimeout(self.timeout, self._protocol.statematcher(connection, RedisConnectionStateEvent.CONNECTION_UP, False)):
                yield m
            if container.timeout:
                raise RedisConnectionDown('Disconnected from redis server')        
    def _get_default_connection(self, container):
        if not self._defaultconn:
            raise RedisConnectionDown('Not connected to redis server')
        if self._lockconnmark is not None:
            if self._lockconnmark >= 0:
                if not self._defaultconn.connected or self._defaultconn.connmark != self._lockconnmark:
                    raise RedisConnectionRestarted('Disconnected from redis server; reconnected is not allowed in with scope')
                else:
                    return
        for m in self._get_connection(container, self._defaultconn):
            yield m
        if self._lockconnmark is not None and self._lockconnmark < 0:
            self._lockconnmark = self._defaultconn.connmark
    def _shutdown_conn(self, container, connection):
        if connection:
            if connection.connected:
                # Send quit
                try:
                    for m in self._protocol.send_command(connection, container, 'QUIT'):
                        yield m
                    for m in container.waitWithTimeout(1, self._protocol.statematcher(connection)):
                        yield m
                except Exception:
                    for m in connection.shutdown():
                        yield m
                else:
                    if container.timeout:
                        for m in connection.shutdown(True):
                            yield m
            else:
                for m in connection.shutdown():
                    yield m
    def shutdown(self, container):
        '''
        Shutdown all connections to Redis server
        '''
        c = self._defaultconn
        self._defaultconn = None
        for m in self._shutdown_conn(container, c):
            yield m
    def release(self, container):
        '''
        Release the connection, leave it to be reused later.
        '''
        if not self._parent:
            for m in self.shutdown(container):
                yield m
        else:
            for m in self._parent._release_conn(container, self._defaultconn):
                yield m
    @contextmanager
    def context(self, container, release = True, lockconn = True):
        '''
        Use with statement to manage the connection
        
        :params release: if True(default), release the connection when leaving with scope
        
        :params lockconn: if True(default), do not allow reconnect during with scope;
                execute commands on a disconnected connection raises Exceptions.
        '''
        try:
            if lockconn:
                if self._lockconnmark is None:
                    # Lock next connmark
                    self._lockconnmark = -1
                    locked = True
            yield self
        finally:
            if locked:
                self._lockconnmark = None
            self._lockconnmark = None
            if release:
                container.subroutine(self.release(container), False)
    @_conn
    def execute_command(self, container, *args):
        '''
        execute command on current connection
        '''
        for m in self._protocol.execute_command(self._defaultconn, container, *args):
            yield m
    @_conn
    def batch_execute(self, container, *cmds):
        '''
        execute a batch of commands on current connection in pipeline mode
        '''
        for m in self._protocol.batch_execute(self._defaultconn, container, *cmds):
            yield m
    def register_script(self, container, script):
        '''
        register a script to this connection.
        
        :returns: registered script. This is a tuple (sha1, script). Pass the tuple to
                  eval_registered, ensure_registerd as registerd_script parameter.
        '''
        if len(script) < 43:
            container.retvalue = (None, script)
        else:
            for m in self.execute_command(container, 'SCRIPT', 'LOAD', script):
                yield m
            container.retvalue = (container.retvalue, script)
    def eval_registered(self, container, registerd_script, *args):
        '''
        eval a registered script. If the script is not cached on the server, it is automatically cached. 
        '''
        if registerd_script[0]:
            try:
                for m in self.execute_command(container, 'EVALSHA', registerd_script[0], *args):
                    yield m
            except RedisReplyException as exc:
                if exc.subtype == 'NOSCRIPT':
                    for m in self.execute_command(container, 'EVAL', registerd_script[1], *args):
                        yield m
                else:
                    raise
        else:
            for m in self.execute_command(container, 'EVAL', registerd_script[1], *args):
                yield m
    def ensure_registerd(self, container, *scripts):
        '''
        Ensure that these scripts are cached on the server. Important when using scripts with batch_execute.
        :param container: routine container.
        :param \*scripts: registered script tuples, return value of register_script 
        '''
        loading = dict((s[0], s[1]) for s in scripts if s[0])
        if loading:
            keys = list(loading.keys())
            for m in self.execute_command(container, 'SCRIPT', 'EXISTS', *keys):
                yield m
            r = container.retvalue
            cmds = [('SCRIPT', 'LOAD', s) for s in (loading[keys[i]] for i in range(0, len(keys)) if not r[i])]
            if cmds:
                for m in self.batch_execute(container, cmds):
                    yield m
    
class RedisClient(RedisClientBase):
    def __init__(self, url = None, db = None, protocol = None):
        '''
        Redis client to communicate with Redis server. Several connections are created for different functions.
         
        :param url: connectiom url, e.g. 'tcp://localhost/'.
                    If not specified, redisclient.url in configuration is used
        
        :param db: default database. If not specified, redisclient.db in configuration is used,
                   which defaults to 0.
        
        :param protocol: use a pre-created protocol instance instead of creating a new instance
        '''
        RedisClientBase.__init__(self, protocol=protocol)
        if url:
            self.url = url
        if db is not None:
            self.db = db
        self._subscribeconn = None
        self._subscribecounter = {}
        self._psubscribecounter = {}
        self._connpool = []
        self._shutdown = False
    def _create_client(self, container):
        if self._shutdown:
            raise IOError('RedisClient already shutdown')
        conn = Client(self.url, self._protocol, container.scheduler,
                                       getattr(self, 'key', None),
                                       getattr(self, 'certificate', None),
                                       getattr(self, 'ca_certs', None))
        conn.start()
        return conn
    def _get_default_connection(self, container):
        if not self._defaultconn:
            self._defaultconn = self._create_client(container)
            for m in RedisClientBase._get_default_connection(self, container):
                yield m
            for m in self._protocol.send_command(self._defaultconn, container, 'SELECT', str(self.db)):
                yield m
        else:
            for m in RedisClientBase._get_default_connection(self, container):
                yield m
    def _get_subscribe_connection(self, container):
        if not self._subscribeconn:
            self._subscribeconn = self._create_client(container)
        for m in RedisClientBase._get_connection(self, container, self._subscribeconn):
            yield m
    def get_connection(self, container):
        '''
        Get an exclusive connection, useful for blocked commands and transactions.
        
        You must call release or shutdown (not recommanded) to return the connection after use.
        
        :param container: routine container
        
        :returns: RedisClientBase object, with some commands same as RedisClient like execute_command,
                  batch_execute, register_script etc.
        '''
        if self._connpool:
            conn = self._connpool.pop()
            container.retvalue = RedisClientBase(conn, self)
        else:
            conn = self._create_client(container)
            for m in RedisClientBase._get_connection(self, container, conn):
                yield m
            for m in self._protocol.send_command(conn, container, 'SELECT', str(self.db)):
                yield m
            container.retvalue = RedisClientBase(conn, self)
    def _release_conn(self, container, connection):
        if connection:
            if self._shutdown or not connection.connected:
                for m in self._shutdown_conn(container, connection):
                    yield m
            else:
                if connection.redis_select != str(self.db):
                    for m in self._protocol.send_command(connection, container, 'SELECT', str(self.db)):
                        yield m
                self._connpool.append(connection)
    def execute_command(self, container, *args):
        '''
        Execute command on Redis server:
          - For (P)SUBSCRIBE/(P)UNSUBSCRIBE, the command is sent to the subscribe connection.
            It is recommended to use (p)subscribe/(p)unsubscribe method instead of directly call the command
          - For BLPOP, BRPOP, BRPOPLPUSH, the command is sent to a separated connection. The connection is
            recycled after command returns.
          - For other commands, the command is sent to the default connection.
        '''
        if args:
            cmd = _str(args[0]).upper()
            if cmd in ('SUBSCRIBE', 'UNSUBSCRIBE', 'PSUBSCRIBE', 'PUNSUBSCRIBE'):
                for m in self._get_subscribe_connection(container):
                    yield m
                for m in self._protocol.execute_command(self._subscribeconn, container, *args):
                    yield m
                return
            elif cmd in ('BLPOP', 'BRPOP', 'BRPOPLPUSH'):
                def _blocked_command():
                    for m in self.get_connection(container):
                        yield m
                    c = container.retvalue
                    with c.context(container):
                        try:
                            for m in c.execute_command(container, *args):
                                yield m
                        except BaseException as exc:
                            for m in c.shutdown(container):
                                yield m
                            raise exc
                with closing(container.delegateOther(_blocked_command(), container, forceclose = True)) as g:
                    for m in g:
                        yield m
                return
        for m in RedisClientBase.execute_command(self, container, *args):
            yield m
    def subscribe(self, container, *keys):
        '''
        Subscribe to specified channels
        :param container: routine container
        :param *keys: subscribed channels
        :returns: list of event matchers for the specified channels
        '''
        for m in self._get_subscribe_connection(container):
            yield m
        realkeys = []
        for k in keys:
            count = self._subscribecounter.get(k, 0)
            if count == 0:
                realkeys.append(k)
            self._subscribecounter[k] = count + 1
        if realkeys:
            for m in self._protocol.execute_command(self._subscribeconn, container, 'SUBSCRIBE', *realkeys):
                yield m
        container.retvalue = [self._protocol.subscribematcher(self._subscribeconn, k, None, RedisSubscribeMessageEvent.MESSAGE) for k in keys]
    def unsubscribe(self, container, *keys):
        '''
        Unsubscribe specified channels. Every subscribed key should be unsubscribed exactly once, even if duplicated subscribed.
        :param container: routine container
        :param *keys: subscribed channels
        '''
        for m in self._get_subscribe_connection(container):
            yield m
        realkeys = []
        for k in keys:
            count = self._subscribecounter.get(k, 0)
            if count <= 1:
                realkeys.append(k)
                try:
                    del self._subscribecounter[k]
                except KeyError:
                    pass
            else:
                self._subscribecounter[k] = count - 1
        if realkeys:
            for m in self._protocol.execute_command(self._subscribeconn, container, 'UNSUBSCRIBE', *realkeys):
                yield m
        container.retvalue = None
    def psubscribe(self, container, *keys):
        '''
        Subscribe to specified globs
        :param container: routine container
        :param *keys: subscribed globs
        :returns: list of event matchers for the specified globs 
        '''
        for m in self._get_subscribe_connection(container):
            yield m
        realkeys = []
        for k in keys:
            count = self._psubscribecounter.get(k, 0)
            if count == 0:
                realkeys.append(k)
            self._psubscribecounter[k] = count + 1
        for m in self._protocol.execute_command(self._subscribeconn, container, 'PSUBSCRIBE', *realkeys):
            yield m
        container.retvalue = [self._protocol.subscribematcher(self._subscribeconn, k, None, RedisSubscribeMessageEvent.PMESSAGE) for k in keys]
    def punsubscribe(self, container, *keys):
        '''
        Unsubscribe specified globs. Every subscribed glob should be unsubscribed exactly once, even if duplicated subscribed.
        
        :param container: routine container
        
        :param *keys: subscribed globs
        '''
        for m in self._get_subscribe_connection(container):
            yield m
        realkeys = []
        for k in keys:
            count = self._subscribecounter.get(k, 0)
            if count == 1:
                realkeys.append(k)
                del self._subscribecounter[k]
            else:
                self._subscribecounter[k] = count - 1
        for m in self._protocol.execute_command(self._subscribeconn, container, 'PUNSUBSCRIBE', *keys):
            yield m
    def shutdown(self, container):
        '''
        Shutdown all connections. Exclusive connections created by get_connection will shutdown after release()
        '''
        p = self._connpool
        self._connpool = []
        self._shutdown = True
        if self._defaultconn:
            p.append(self._defaultconn)
            self._defaultconn = None
        if self._subscribeconn:
            p.append(self._subscribeconn)
            self._subscribeconn = None
        for m in container.executeAll([self._shutdown_conn(container, o)
                                       for o in p]):
            yield  m
    class _RedisConnection(object):
        def __init__(self, client, container):
            self._client = client
            self._container = container
        def start(self, asyncstart = False):
            pass
        def shutdown(self):
            if self._client:
                try:
                    for m in self._container.delegate(self._client.shutdown(self._container), True):
                        yield m
                finally:
                    self._client = None
                    self._container = None
    def make_connobj(self, container):
        '''
        Return an object to be used like a connection. Put the connection-like object in module.connections
        to make RedisClient shutdown on module unloading.
        '''
        return self._RedisConnection(self, container)
    def subscribe_state_matcher(self, container, connected = True):
        '''
        Return a matcher to match the subscribe connection status.
        
        :param container: a routine container. NOTICE: this method is not a routine.
        
        :param connected: if True, the matcher matches connection up. If False, the matcher matches
               connection down.
        
        :returns: an event matcher.
        '''
        if not self._subscribeconn:
            self._subscribeconn = self._create_client(container)
        return RedisConnectionStateEvent.createMatcher(
                    RedisConnectionStateEvent.CONNECTION_UP if connected else RedisConnectionStateEvent.CONNECTION_DOWN,
                    self._subscribeconn
                    )
        


import logging
import copy

from uuid import uuid1
from vlcp.server.module import Module,depend,api,callAPI
from vlcp.event.runnable import RoutineContainer
from vlcp.utils.networkmodel import *
from vlcp.utils.dataobject import watch_context,set_new,dump,ReferenceObject
from vlcp.utils.ethernet import ip4_addr
from vlcp.utils.netutils import format_network_cidr,check_ip_address, parse_ip4_network, ip_in_network

import vlcp.service.kvdb.objectdb as objectdb
from vlcp.config.config import defaultconfig

logger = logging.getLogger("vrouterapi")


@depend(objectdb.ObjectDB)
@defaultconfig
class VRouterApi(Module):
    """
    Standard network model for L3 SDN
    """
    def __init__(self,server):
        super(VRouterApi,self).__init__(server)

        self.app_routine = RoutineContainer(self.scheduler)
        self.app_routine.main = self.main
        self.routines.append(self.app_routine)
        self._reqid = 0

        self.createAPI(api(self.createvirtualrouter,self.app_routine),
                       api(self.createvirtualrouters,self.app_routine),
                       api(self.updatevirtualrouter,self.app_routine),
                       api(self.updatevirtualrouters,self.app_routine),
                       api(self.deletevirtualrouter,self.app_routine),
                       api(self.deletevirtualrouters,self.app_routine),
                       api(self.listvirtualrouters,self.app_routine),
                       api(self.addrouterinterface,self.app_routine),
                       api(self.addrouterinterfaces,self.app_routine),
                       api(self.removerouterinterface,self.app_routine),
                       api(self.removerouterinterfaces,self.app_routine),
                       api(self.listrouterinterfaces,self.app_routine))
    def main(self):
        logger.info(" viperflow VRouterApi running --")
        
        if False:
            yield 
    def _dumpkeys(self,keys):
        self._reqid += 1
        reqid = ('virtualrouter',self._reqid)

        for m in callAPI(self.app_routine,'objectdb','mget',{'keys':keys,'requestid':reqid}):
            yield m

        retobjs = self.app_routine.retvalue

        with watch_context(keys,retobjs,reqid,self.app_routine):
            self.app_routine.retvalue = [dump(v) for v in retobjs]

    def _getkeys(self,keys):
        self._reqid += 1
        reqid = ('virtualrouter',self._reqid)
        for m in callAPI(self.app_routine,'objectdb','mget',{'keys':keys,'requestid':reqid}):
            yield m
        with watch_context(keys,self.app_routine.retvalue,reqid,self.app_routine):
            pass

    def createvirtualrouter(self,id=None,**kwargs):
        """
        Create a virtual router
        
        :param id: Virtual router id. If omitted, an UUID is generated.
        
        :param \*\*kwargs: extra attributes for creation.
        
        :return: A dictionary of information of the virtual router.
        """
        if not id:
            id = str(uuid1())

        router = {"id":id}
        router.update(kwargs)

        for m in self.createvirtualrouters([router]):
            yield m

    def createvirtualrouters(self,routers):
        """
        Create multiple virtual routers in a transaction
        """
        idset = set()
        newrouters = []
        for router in routers:
            router = copy.deepcopy(router)
            if "id" not in router:
                router["id"] = str(uuid1())
            else:
                if router["id"] in idset:
                    raise ValueError("id repeat " + router['id'])
                else:
                    idset.add(router['id'])

            newrouters.append(router)
        
        routerkeys = [VRouter.default_key(r['id']) for r in newrouters]
        routersetkey = [VRouterSet.default_key()]

        routerobjects = [self._createvirtualrouter(**r) for r in newrouters ]

        def createrouterdb(keys,values):
            routerset = values[0]

            for i in range(0,len(routerkeys)):
                values[i+1] = set_new(values[i+1],routerobjects[i])
                routerset.set.dataset().add(routerobjects[i].create_weakreference())

            return keys,values
        try:
            for m in callAPI(self.app_routine,"objectdb","transact",
                             {"keys":routersetkey+routerkeys,"updater":createrouterdb}):
                yield m
        except:
            raise
        else:
            for m in self._dumpkeys(routerkeys):
                yield m

    def _createvirtualrouter(self,id,**kwargs):

        router = VRouter.create_instance(id)

        # [(prefix, nexthop),(prefix, nexthop)]
        if 'routes' in kwargs:
            for e in kwargs['routes']:
                ip_prefix = e[0]
                nexthop = e[1]

                if ip_prefix and nexthop:
                    ip_prefix = format_network_cidr(ip_prefix)
                    nexthop = check_ip_address(nexthop)
                    router.routes.append((ip_prefix,nexthop))
                else:
                    raise ValueError(" routes format error " + e)

        for k,v in kwargs.items():
            if k != "routes":
                setattr(router,k,v)

        return router

    def updatevirtualrouter(self,id,**kwargs):
        """
        Update virtual router
        """
        if not id:
            raise ValueError(" must specify id")
        router = {"id":id}
        router.update(kwargs)

        for m in self.updatevirtualrouters([router]):
            yield m

    def updatevirtualrouters(self,routers):
        "Update multiple virtual routers"
        idset = set()
        for router in routers:
            if 'id' not in router:
                raise ValueError(" must specify id")
            else:
                if router['id'] in idset:
                    raise ValueError(" id repeat " + router['id'])
                else:
                    idset.add(router['id'])

            # if routers updating ,, check format
            if 'routes' in router:
                for r in router['routes']:
                    ip_prefix = r[0]
                    nexthop = r[1]

                    if ip_prefix and nexthop:
                        ip_prefix = parse_ip4_network(ip_prefix)
                        nexthop = check_ip_address(nexthop)
                    else:
                        raise ValueError("routes format error " + r)

        routerkeys = [VRouter.default_key(r['id']) for r in routers]

        def updaterouter(keys,values):

            for i in range(0,len(routers)):
                if values[i]:
                    for k,v in routers[i].items():

                        if k == 'routes':
                            # update routers accord new static routes
                            values[i].routes = []
                            for pn in v:
                                values[i].routes.append(pn)
                        else:
                            setattr(values[i],k,v)
                else:
                    raise ValueError("router object not exists " + routers[i]['id'])
            
            return keys,values
        try:
            for m in callAPI(self.app_routine,'objectdb','transact',
                             {'keys':routerkeys,'updater':updaterouter}):
                yield m
        except:
            raise
        else:

            for m in self._dumpkeys(routerkeys):
                yield m

    def deletevirtualrouter(self,id):
        "Delete virtual router"
        if not id:
            raise ValueError("must specify id")

        router = {"id":id}

        for m in self.deletevirtualrouters([router]):
            yield m

    def deletevirtualrouters(self,routers):
        "Delete multiple virtual routers"
        idset = set()
        for router in routers:
            if 'id' not in router:
                raise ValueError(" must specify id")
            else:
                if router['id'] in idset:
                    raise ValueError(" id repeat " + router['id'])
                else:
                    idset.add(router['id'])

        routersetkey = [VRouterSet.default_key()]
        routerkeys = [VRouter.default_key(v['id']) for v in routers]

        def deletevrouter(keys,values):
            routerset = values[0]

            for i in range(0,len(routers)):
                if values[i+1].interfaces.dataset():
                    raise ValueError("there are still interface(s) in this router, delete them first")
                routerset.set.dataset().discard(values[i+1].create_weakreference())

            return keys,[routerset] + [None] * len(routers)
        try:
            for m in callAPI(self.app_routine,"objectdb","transact",
                             {"keys":routersetkey + routerkeys,"updater":deletevrouter}):
                yield m
        except:
            raise
        else:
            self.app_routine.retvalue = {"status":"OK"}

    def listvirtualrouters(self,id=None,**kwargs):
        """
        Query virtual router
        
        :param id: if specified, only return virtual router with this ID
        
        :param \*\*kwargs: customized filter
        
        :return: a list of dictionaries each stands for a matched virtual router.
        """
        if id:
            routerkey = [VRouter.default_key(id)]

            for m in self._getkeys(routerkey):
                yield m

            retobj = self.app_routine.retvalue

            if all(getattr(retobj,k,None) == v for k,v in kwargs.items()):
                self.app_routine.retvalue = dump(retobj)
            else:
                self.app_routine.retvalue = []
        else:
            # we get all router from vrouterset index
            routersetkey = VRouterSet.default_key()

            self._reqid = +1
            reqid = ("virtualrouter",self._reqid)

            def set_walker(key,set,walk,save):

                for weakobj in set.dataset():
                    routerkey = weakobj.getkey()

                    try:
                        router = walk(routerkey)
                    except KeyError:
                        pass
                    else:
                        if all(getattr(router,k,None) == v for k,v in kwargs.items()):
                            save(routerkey)

            def walker_func(set_func):
                def walker(key,obj,walk,save):
                    
                    if obj is None:
                        return
                    
                    set_walker(key,set_func(obj),walk,save)

                return walker

            for m in callAPI(self.app_routine,"objectdb","walk",
                             {"keys":[routersetkey],"walkerdict":{routersetkey:walker_func(lambda x:x.set)},
                              "requestid":reqid}):
                yield m

            keys ,values = self.app_routine.retvalue

            with watch_context(keys,values,reqid,self.app_routine):
                self.app_routine.retvalue = [dump(r) for r in values]

    def addrouterinterface(self,router,subnet,id=None,**kwargs):
        """
        Connect virtual router to a subnet
        
        :param router: virtual router ID
        
        :param subnet: subnet ID
        
        :param id: router port ID
        
        :param \*\*kwargs: customized options
        
        :return: A dictionary of information of the created router port
        """
        if not id:
            id = str(uuid1())

        if not router:
            raise ValueError(" must specify routerid")

        if not subnet:
            raise ValueError(" must specify subnetid")

        interface = {"id":id,"router":router,"subnet":subnet}

        interface.update(kwargs)
        for m in self.addrouterinterfaces([interface]):
            yield m

    def addrouterinterfaces(self,interfaces):
        """
        Create multiple router interfaces
        """
        idset = set()
        newinterfaces = []
        for interface in interfaces:
            interface = copy.deepcopy(interface)
            if 'id' in interface:
                if interface['id'] not in idset:
                    idset.add(interface['id'])
                else:
                    raise ValueError(" id repeat "+interface['id'])
            else:
                interface['id'] = str(uuid1())

            if 'router' not in interface:
                raise ValueError("must specify router id")

            if 'subnet' not in interface:
                raise ValueError("must specify subnet id")

            newinterfaces.append(interface)

        routerportkeys = [RouterPort.default_key(interface['id'])
                           for interface in newinterfaces]
        routerportobjects = [self._createrouterport(**interface) for interface in newinterfaces]

        routerkeys = list(set([routerport.router.getkey() for routerport in routerportobjects]))

        subnetkeys = list(set([routerport.subnet.getkey() for routerport in routerportobjects]))

        subnetmapkeys = [SubNetMap.default_key(SubNet._getIndices(k)[1][0]) for k in subnetkeys]
        keys = routerkeys + subnetkeys + subnetmapkeys + routerportkeys
        newrouterportdict = dict(zip(routerportkeys,routerportobjects))

        def addrouterinterface(keys,values):

            rkeys = keys[0:len(routerkeys)]
            robjs = values[0:len(routerkeys)]

            snkeys = keys[len(routerkeys):len(routerkeys) + len(subnetkeys)]
            snobjs = values[len(routerkeys):len(routerkeys) + len(subnetkeys)]

            snmkeys = keys[len(routerkeys) + len(subnetkeys):len(routerkeys) + len(subnetkeys) + len(subnetmapkeys)]
            snmobjs = values[len(routerkeys) + len(subnetkeys):len(routerkeys) + len(subnetkeys) + len(subnetmapkeys)]

            rpkeys = keys[len(routerkeys) + len(subnetkeys) + len(subnetmapkeys):]
            rpobjs = values[len(routerkeys) + len(subnetkeys) + len(subnetmapkeys):]

            rdict = dict(zip(rkeys,robjs))
            sndict = dict(zip(snkeys,zip(snobjs,snmobjs)))
            rpdict = dict(zip(rpkeys,rpobjs))
            
            for i,interface in enumerate(newinterfaces):
                routerport = interface['id']
                router = interface['router']
                subnet = interface['subnet']

                routerobj = rdict.get(VRouter.default_key(router))
                subnetobj,subnetmapobj = sndict.get(SubNet.default_key(subnet))
                newrouterport = newrouterportdict.get(RouterPort.default_key(routerport))
                routerport = rpdict.get(RouterPort.default_key(routerport))
                
                if routerobj and subnetobj and subnetmapobj:

                    # now subnet only have one router ,,so check it
                    if not hasattr(subnetobj,'router'):

                        # new router port special ip address , we check it in subnetmap
                        if hasattr(newrouterport,"ip_address"):
                            ipaddress = ip4_addr(newrouterport.ip_address)

                            n,p = parse_ip4_network(subnetobj.cidr)

                            if not ip_in_network(ipaddress,n,p):
                                raise ValueError(" special ip address not in subnet cidr")

                            if str(ipaddress) not in subnetmapobj.allocated_ips:
                                subnetmapobj.allocated_ips[str(ipaddress)] = newrouterport.create_weakreference()
                            else:
                                raise ValueError(" ip address have used in subnet "+ newrouterport.ip_address)
                        else:
                            # not have special ip address, special gateway to this only router port
                            # gateway in subnet existed be sure

                            #
                            # when this case , special subnet gateway as ip_address
                            # but we do not set ip_address attr ,  when subnet gateway change ,
                            # we do not need to change router port attr

                            #setattr(newrouterport,"ip_address",subnetobj.gateway)

                            # it may be subnet not have gateway, checkout it
                            if not hasattr(subnetobj,"gateway"):
                                raise ValueError(" interface not special ip_address and subnet has no gateway")

                        #routerport = set_new(routerport,newrouterport)
                        values[len(routerkeys) + len(subnetkeys) + len(subnetmapkeys)+i] = set_new(
                                    values[len(routerkeys) + len(subnetkeys) + len(subnetmapkeys) + i],
                                    routerportobjects[i]
                                )
                        subnetobj.router = newrouterport.create_weakreference()
                        routerobj.interfaces.dataset().add(newrouterport.create_weakreference())
                    else:
                        raise ValueError(" subnet " + subnet + " have router port " + subnetobj.router.getkey())
                else:
                    raise ValueError(" routerobj " + router + "or subnetobj " + subnet + " not existed ")

            return keys,values
        try:
            for m in callAPI(self.app_routine,'objectdb','transact',
                         {"keys":keys,"updater":addrouterinterface}):
                yield m
        except:
            raise
        else:
            for m in self._dumpkeys(routerportkeys):
                yield m

    def _createrouterport(self,id,router,subnet,**kwargs):
        routerport = RouterPort.create_instance(id)
        routerport.router = ReferenceObject(VRouter.default_key(router))
        routerport.subnet = ReferenceObject(SubNet.default_key(subnet))

        for k,v in kwargs.items():
            setattr(routerport,k,v)

        return routerport

    def removerouterinterface(self,router,subnet):
        """
        Remote a subnet from the router
        
        :param router: virtual router ID
        
        :param subnet: subnet ID
        
        :return: ``{"status": "OK"}``
        """
        if not router:
            raise ValueError("must specify router id")

        if not subnet:
            raise ValueError("must specify subnet id")

        interface = {"router":router,"subnet":subnet}

        for m in self.removerouterinterfaces([interface]):
            yield m

    def removerouterinterfaces(self,interfaces):
        """
        Remote multiple subnets from routers
        """

        # idset use to check repeat subnet!
        idset = set()
        delete_interfaces = []

        for interface in interfaces:
            interface = copy.deepcopy(interface)
            if "router" not in interface:
                raise ValueError(" must specify router=id")
            if "subnet" not in interface:
                raise ValueError(" must specify subnet=id")
            else:
                if interface["subnet"] in idset:
                    raise ValueError("repeated subnet id " + interface["subnet"])
                else:
                    idset.add(interface["subnet"])

            delete_interfaces.append(interface)

        subnetkeys = [SubNet.default_key(interface["subnet"]) for interface in delete_interfaces]
        subnetmapkeys = [SubNetMap.default_key(interface["subnet"]) for interface in delete_interfaces]
        for m in self._getkeys(subnetkeys):
            yield m

        subnetobjs = self.app_routine.retvalue

        if None in subnetobjs:
            raise ValueError(" subnet object not existed " +\
                             SubNet._getIndices(subnetkeys[subnetobjs.index(None)])[1][0])

        sndict = dict(zip(subnetkeys,subnetobjs))

        for interface in delete_interfaces:
            if SubNet.default_key(interface['subnet']) in sndict:
                snobj = sndict[SubNet.default_key(interface["subnet"])]
                if hasattr(snobj,"router"):
                    routerportid = RouterPort._getIndices(snobj.router.getkey())[1][0]
                    interface['routerport'] = routerportid
                else:
                    raise ValueError("subnet " + interface["subnet"] + " is not plugged to router" )


        subnetkeys = list(set(subnetkeys))
        
        routerportkeys = [s.router.getkey() for s in subnetobjs]
        
        routerportkeys = list(set(routerportkeys))
        
        routerkeys = [VRouter.default_key(interface["router"]) for interface in delete_interfaces]

        routerkeys = list(set(routerkeys))
        
        def delete_interface(keys,values):
            rkeys = keys[0:len(routerkeys)]
            robjs = values[0:len(routerkeys)]

            snkeys = keys[len(routerkeys):len(routerkeys) + len(subnetkeys)]
            snobjs = values[len(routerkeys):len(routerkeys) + len(subnetkeys)]

            snmkeys = keys[len(routerkeys) + len(subnetkeys):len(routerkeys) + len(subnetkeys) + len(subnetmapkeys)]
            snmobjs = values[len(routerkeys) + len(subnetkeys):len(routerkeys) + len(subnetkeys) + len(subnetmapkeys)]

            rpkeys = keys[-len(routerportkeys):]
            rpobjs = values[-len(routerportkeys):]

            rdict = dict(zip(rkeys,robjs))
            sndict = dict(zip(snkeys,snobjs))
            snmdict = dict(zip(snmkeys,snmobjs))
            rpdict = dict(zip(rpkeys,rpobjs))

            for interface in delete_interfaces:
                r = rdict.get(VRouter.default_key(interface["router"]),None)
                sn = sndict.get(SubNet.default_key(interface["subnet"]),None)
                snm = snmdict.get(SubNetMap.default_key(interface["subnet"]),None)
                rp = rpdict.get(RouterPort.default_key(interface["routerport"]),None)

                if r and sn and snm and rp:
                    # has not attr ip_address, means router port use subnet gateway as ip_address
                    if hasattr(rp,'ip_address'):
                        # it means have allocated ip address in subnetmap, delete it
                        ipaddress = ip4_addr(rp.ip_address)
                        del snm.allocated_ips[str(ipaddress)]

                    # one subnet only have one router , so del this attr
                    if hasattr(sn,'router'):
                        delattr(sn,'router')

                    if rp.create_weakreference() in r.interfaces.dataset():
                        r.interfaces.dataset().discard(rp.create_weakreference())
                    else:
                        raise ValueError("router " + interface["router"] + " have no router port " +
                                         interface["routerport"])
                else:
                    raise ValueError("route " + interface["router"] + " subnet " + interface["subnet"] +
                                     " routerport " + interface["routerport"] + " not existed!")

            return keys,robjs + snobjs + snmobjs + [None]*len(routerportkeys)

        transact_keys = routerkeys + subnetkeys + subnetmapkeys + routerportkeys

        for m in callAPI(self.app_routine,"objectdb","transact",{"keys":transact_keys,
                                                                 "updater":delete_interface}):
            yield m

        self.app_routine.retvalue = {'status': 'OK'}

    def listrouterinterfaces(self,id,**kwargs):
        """
        Query router ports from a virtual router
        
        :param id: virtual router ID
        
        :param \*\*kwargs: customized filters on router interfaces
        
        :return: a list of dictionaries each stands for a matched router interface
        """
        if not id:
            raise ValueError(" must special router id")

        routerkey = VRouter.default_key(id)

        self._reqid = +1
        reqid = ("virtualrouter",self._reqid)
        
        def set_walker(key,interfaces,walk,save):

            for weakobj in interfaces.dataset():
                vpkey = weakobj.getkey()
                
                try:
                    virtualport = walk(vpkey)
                except KeyError:
                    pass
                else:
                    if all(getattr(virtualport,k,None) == v for k,v in kwargs.items()):
                        save(vpkey)
        
        def walk_func(filter_func):

            def walker(key,obj,walk,save):
                if obj is None:
                    return
            
                set_walker(key,filter_func(obj),walk,save)

            return walker
        for m in callAPI(self.app_routine,"objectdb","walk",
                {"keys":[routerkey],"walkerdict":{routerkey:walk_func(lambda x:x.interfaces)},
                    "requestid":reqid}):
            yield m

        keys,values = self.app_routine.retvalue

        with watch_context(keys,values,reqid,self.app_routine):
            self.app_routine.retvalue = [dump(r) for r in values]

    def load(self,container):

        initkeys = [VRouterSet.default_key(),DVRouterForwardSet.default_key(),
                    DVRouterExternalAddressInfo.default_key()]

        def init(keys,values):
            
            if values[0] is None:
                values[0] = VRouterSet()

            if values[1] is None:
                values[1] = DVRouterForwardSet()

            if values[2] is None:
                values[2] = DVRouterExternalAddressInfo()
            
            return keys,values

        
        for m in callAPI(container,"objectdb","transact",
                         {"keys":initkeys,"updater":init}):
            yield m
        
        for m in Module.load(self,container):
            yield m


